using System;
using System.IO;
using System.Collections;
using System.Windows.Forms;
using System.Text;
using MIF.Primitives;
using System.Drawing;
using MyUtils;
using MIFFile;

/// <summary>
/// Summary description for Class1
/// </summary>
namespace MIF
{
    public class MIFTypes
    {
        public const string Character       =   "CHAR";
        public const string Integer         =   "INTEGER";
        public const string SmallInteger    =   "SMALL INTEGER";
        public const string Float           =   "FLOAT";
        public const string Decimal         =   "DECIMAL";
        public const string Date            =   "DATE";
        public const string Logical         =   "LOGICAL";
        public static bool IsCharacter(string aString,int aWidth)
        {
            return (aString.Length <= aWidth);
        }
        public static bool IsInteger(string aString)
        {
            try
            {
                long.Parse(aString);
                return true;
            }
            catch
            {
                return false;
            }
        }
        public static bool IsSmallInteger(string aString)
        {
            try
            {
                int.Parse(aString);
                return true;
            }
            catch
            {
                return false;
            }
        }
        public static bool IsFloat(string aString)
        {
           try
            {
                float.Parse(aString);
                return true;
            }
           catch
            {
                return false;
            }
        }
        public static bool IsDecimal(string aString, int aWidth, int aDecimals)
        {
            if (aString.Trim(new char[] { ',', '.' }).Length > aWidth) return false;
            string[] tokens = Strings.Split(aString, new char[] { ',', '.' });
            if (tokens.Length > 1)
                if (tokens[1].Length > aDecimals) return false;
            return true;
        }
        public static bool IsDate(string aString)
        {
            try
            {
                DateTime.Parse(aString);
                return true;
            }
            catch
            {
                return false;
            }
        }
        public static bool IsLogical(string aString)
        {
            if ((aString != "F") || (aString == "T")) return false;
            return true;
        }
    }
    
    
    public class MIFWords
    {
         public const string Version = "Version";
         public const string CharSet = "Charset";
         public const string Delimiter = "Delimiter";
         public const string Unique = "Unique";
         public const string Index = "Index";
         public const string CoordSys = "CoordSys";
         public const string Transform = "Transform";
         public const string Columns = "Columns";
         public const string Data = "Data";
         public const string None = "none";
         public const string Point = "Point";
         public const string Line = "Line";
         public const string Pline = "Pline";
         public const string Region = "Region";
         public const string Arc = "Arc";
         public const string Text = "Text";
         public const string Rect = "Rect";
         public const string RoundRect = "RoundRect";
         public const string Ellipse = "Ellipse";
         public const string MultiPoint = "MultiPoint";
         public const string Collection = "Collection";
         public const string Symbol = "Symbol";
         public const string Pen = "Pen";
         public const string Smooth = "Smooth";
         public const string Brush = "Brush";
         public const string Center = "Center";
         public const string Font = "Font";
         public const string Multiple = "Multiple"; 
    }
     class Defaults
    {
        public const int    Header_Version = 300;
        public const string Header_Charset = "WindowsCyrillic";
        public const string Header_Delimiter = "\t";
        public const int    Symbol_Shape     = 67;
        public const int    Symbol_Color = 0x000000;
        public const int    Symbol_Size      = 15;
        public const int    Pen_width        = 1;
        public const int    Pen_pattern      = 2;
        public const int    Pen_color        = 0x000000;
        public const int    Brush_pattern    = 2;
        public const int    Brush_forecolor  = 0xFFFFFF;
        public const int    Brush_backcolor  = 0xFFFFFF;
        public const int    Line_width = 1             ;
        public const int    Line_color = 0x000000      ;
        public static readonly Color SelectColor = Color.LimeGreen;
        public const int DistToObjectToSelectInPixels = 5;
        
    }
       
     public class MIFFile
     {

         #region fields
         public readonly string mMifFileName;
         public readonly string mMidFileName;
         public readonly string mGpsFileName;

         public MIFHeader mMIFHeader;
         public MIFData mMIFData;
         public MIDData mMIDData;
         public BaseObject this[int index]
         {
              get {return mMIFData[index];} 
         }
          
         #endregion

         #region methods
         public MIFFile(string aFileName)
         {
             mMIFHeader = new MIFHeader();
             mMIFData = new MIFData();
             mMIDData = new MIDData();
             mMifFileName = aFileName;
             mMidFileName = aFileName.Substring(0, aFileName.IndexOf('.')) + ".MID";
             mGpsFileName = aFileName.Substring(0, aFileName.IndexOf('.')) + ".GPS";
         }

         private bool SaveGPS(string aFileName)
         {
             try
             {
                 using (StreamWriter cr = new StreamWriter(aFileName,false,Encoding.ASCII))
                 {
                     for (int i = 0; i < mMIFData.Objects.Count; i++)
                     {
                         for (int j = 0; j < (mMIFData.Objects[i] as BaseObject).GPSRawStrings.Count; j++)
                         {
                             if (Strings.IsNullOrEmpty((mMIFData.Objects[i] as BaseObject).GPSRawStrings[j] as String)) continue;
                             cr.WriteLine(string.Format("{0}. {1}", i, (mMIFData.Objects[i] as BaseObject).GPSRawStrings[j]));
                         }
                         
                     }
                 }
             }
             catch
             {
                 return false;
             }
             return true;
         }

         public void AddObject(BaseObject aObject)
         {
             
             mMIFData.Objects.Add(aObject);
             ObjectColumns columns = new ObjectColumns(1);
             aObject.Properties = columns;
             for (int i = 0; i < mMIFHeader.mColumns.Length;i++ )
                 columns.Add(new ObjectColumn(mMIFHeader.mColumns[i].mName, "", mMIFHeader.mColumns[i].mType));
             mMIDData.mRows.Add(columns);
         }

         public void DrawPoint(Styles aStyle,Point aMouseDownPoint,PictureBox aPictureBox,string aGPSRawString)
         {
             PointDouble point = MapTrans.GetDegree(aMouseDownPoint);
             MIFSymbol symbol = new MIFSymbol(aStyle.SymbolID, Defaults.Symbol_Color, aStyle.SymbolSize);
             MIFPoint mifpoint = new MIFPoint(point.X, point.Y, symbol);
             mifpoint.GPSRawStrings.Add(aGPSRawString);
             AddObject(mifpoint);
             mifpoint.Draw(aPictureBox.Image as Bitmap);
             aPictureBox.Invalidate();
         }

         public void DrawLine(Styles aStyle,Point aMouseDownPoint1,Point aMouseDownPoint2,PictureBox aPictureBox,ArrayList aGpsRawStrings)
         {
             PointDouble point1 = MapTrans.GetDegree(aMouseDownPoint1);
             PointDouble point2 = MapTrans.GetDegree(aMouseDownPoint2);
             MIFNode node1 = new MIFNode(point1.X, point1.Y,true);
             MIFNode node2= new MIFNode(point2.X, point2.Y,true);
             MIFPen mifpen = new MIFPen(aStyle.LineWidth,Defaults.Pen_pattern,aStyle.LineColor);
             MIFLine mifline = new MIFLine(node1,node2,mifpen);
             mifline.GPSRawStrings.AddRange(aGpsRawStrings);
             AddObject(mifline);
             mifline.Draw(aPictureBox.Image as Bitmap);
             aPictureBox.Invalidate();
         }

         public void DrawPLine(ArrayList aPoints, int aWidth, int aColor, PictureBox aPictureBox)
         {
             ArrayList arr_nodes = new ArrayList(aPoints.Count);
             for (int i = 0; i < aPoints.Count; i++)
             {
                 arr_nodes.Add(new MIFNode(MapTrans.GetDegree((Point)aPoints[i]).X,MapTrans.GetDegree((Point)aPoints[i]).Y,true));
             }
             MIFSection section  = new MIFSection(arr_nodes,SectionType.PolyLine,null,null);
             
             ArrayList  sections = new ArrayList(1);
             sections.Add(section);
             section.MIFPen = new MIFPen(aWidth, 0, aColor);
             MIFPolyLine mif_pline = new MIFPolyLine(false, section.MIFPen, sections);
             AddObject(mif_pline);
             mif_pline.Draw(aPictureBox.Image as Bitmap);
             aPictureBox.Invalidate();
         }
         public void Select(Point aMouseDownPoint,MouseEventArgs e,KeyEventArgs k,PictureBox aPictureBox,ArrayList aSelectedObjects)
         {
                aSelectedObjects.Clear();
                PointDouble deg_point = MapTrans.GetDegree(aMouseDownPoint);
                object current_object=null;
                ArrayList selected_objects = new ArrayList();

                for (int i = 0; i < mMIFData.Objects.Count; i++)
                {
                    BaseObject o = mMIFData.Objects[i] as BaseObject;
                    if (o.DoSelect(deg_point.X, deg_point.Y, Defaults.DistToObjectToSelectInPixels))
                    {
                        selected_objects.Add(o);
                    }
                    
                }


                if (frmChooseSelectedObject.Run(selected_objects, ref current_object))
                {
                    if ((e!=null) && (e.Button ==MouseButtons.Left))
                    {
                        
                        (current_object as BaseObject).IsSelected = !(current_object as BaseObject).IsSelected;
                        (current_object as BaseObject).Draw(aPictureBox.Image as Bitmap);    
                        for (int i = 0; i < mMIFData.Objects.Count; i++)
                            {
                                if (!mMIFData[i].IsSelected) continue;
                                if (mMIFData[i].Equals(current_object)) continue;
                                mMIFData[i].IsSelected = false;
                                mMIFData[i].Draw(aPictureBox.Image as Bitmap);
                                break;
                            }

                        
                        if ((current_object as BaseObject).IsSelected)
                            aSelectedObjects.Add(current_object);
                        (current_object as BaseObject).Draw(aPictureBox.Image as Bitmap); 
                        aPictureBox.Invalidate();
                    }
      
                }
                
                return;
            }
         public void Move(Point aMousePoint1, Point aMousePoint2, PictureBox aPictureBox)
         {
             PointDouble p1, p2;
             p1 = MapTrans.GetDegree(aMousePoint1);
             p2 = MapTrans.GetDegree(aMousePoint2);
             PointDouble delta = new PointDouble(p2.X - p1.X, p2.Y - p1.Y);
             MapTrans.Move(delta);
             Draw(aPictureBox.Image as Bitmap);
             aPictureBox.Invalidate();
         }
         public void Zoom(int aKoef,Point aMousePoint,PictureBox aPictureBox)
         {
             PointDouble basepoint = MapTrans.GetDegree(aMousePoint);
             MapTrans.Zoom(aKoef, basepoint);
             Draw(aPictureBox.Image as Bitmap);
             aPictureBox.Invalidate();
             
         }
         public bool DrawInit(PictureBox aPictureBox)
         {
             if (aPictureBox == null) return false;
             if (aPictureBox.Image == null) return false;
             PointDouble base_point = new PointDouble();
             base_point.X = this.mMIFData.GetMinX() + (this.mMIFData.GetMaxX() - this.mMIFData.GetMinX()) / 2;
             base_point.Y = this.mMIFData.GetMinY() + (this.mMIFData.GetMaxY() - this.mMIFData.GetMinY()) / 2;
             MapTrans.Init(base_point, this.mMIFData.MapSize, aPictureBox.Image.Size);
             if (!MapTrans.IsInit) return false;
             Draw(aPictureBox.Image as Bitmap);
             aPictureBox.Invalidate();
             return true;
         }

         public void Clear()
         {
             mMIDData.Clear();
             mMIFData.Clear();
         }

         public bool Read()
         {
             try
             {
                 using (CommandReader mif_reader = new CommandReader(mMifFileName, Encoding.Default))
                 {
                     if (!mMIFHeader.Read(mif_reader)) return false;
                     using (CommandReader mid_reader = new CommandReader(mMidFileName, mMIFHeader.GetEncoding()))
                     {
                         Clear();
                         string line = mif_reader.ReadLine();
                         
                         while (line != null)
                         {
                             line = line.Trim();
                             if (line.Length <= 0) { line = mif_reader.ReadLine(); continue; }
                             string[] words = line.Split(new char[] { ' ', '\t', '\n', '\r' });
                             string word = words[0];
                             if (string.Compare(word, MIFWords.Point, true) == 0)
                             {
                                 MIF.Primitives.MIFPoint obj = MIF.Primitives.MIFPoint.Parse(mMIFData.GetStrObject(mif_reader, line, out line));
                                 mMIFData.Objects.Add(obj);
                                 string l_line = mid_reader.ReadLine();
                                 ObjectColumns row;
                                 if (l_line == null)
                                 {
                                     row = new ObjectColumns(mMIFHeader.mColumns.Length);
                                     row.InitFromHeader(mMIFHeader);
                                 }
                                 else
                                     row = ObjectColumns.Parse(l_line, mMIFHeader);
                                 mMIDData.mRows.Add(row);
                                 obj.Properties = mMIDData[mMIFData.Objects.Count - 1];

                                 continue;

                             }
                             else
                                 if (string.Compare(word, MIFWords.Line, true) == 0)
                                 {
                                     MIF.Primitives.MIFLine obj = MIF.Primitives.MIFLine.Parse(mMIFData.GetStrObject(mif_reader, line, out line));
                                     mMIFData.Objects.Add(obj);

                                     string l_line = mid_reader.ReadLine();
                                     ObjectColumns row;
                                     if (l_line == null)
                                     {
                                         row = new ObjectColumns(mMIFHeader.mColumns.Length);
                                         row.InitFromHeader(mMIFHeader);
                                     }
                                     else
                                         row = ObjectColumns.Parse(l_line, mMIFHeader);
                                     mMIDData.mRows.Add(row);
                                     obj.Properties = mMIDData[mMIFData.Objects.Count - 1];
                                     continue;
                                 }
                                 else
                                     if (string.Compare(word, MIFWords.Pline, true) == 0)
                                     {
                                         MIF.Primitives.MIFPolyLine obj = MIF.Primitives.MIFPolyLine.Parse(mMIFData.GetStrObject(mif_reader, line, out line));
                                         mMIFData.Objects.Add(obj);
                                         string l_line = mid_reader.ReadLine();
                                         ObjectColumns row;
                                         if (l_line == null)
                                         {
                                             row = new ObjectColumns(mMIFHeader.mColumns.Length);
                                             row.InitFromHeader(mMIFHeader);
                                         }
                                         else
                                             row = ObjectColumns.Parse(l_line, mMIFHeader);
                                         mMIDData.mRows.Add(row);
                                         obj.Properties = mMIDData[mMIFData.Objects.Count - 1];
                                         continue;
                                     }
                                     else
                                         if (string.Compare(word, MIFWords.Region, true) == 0)
                                         {

                                             MIF.Primitives.MIFRegion obj = MIF.Primitives.MIFRegion.Parse(mMIFData.GetStrObject(mif_reader, line, out line));
                                             mMIFData.Objects.Add(obj);
                                             string l_line = mid_reader.ReadLine();
                                             ObjectColumns row;
                                             if (l_line == null)
                                             {
                                                 row = new ObjectColumns(mMIFHeader.mColumns.Length);
                                                 row.InitFromHeader(mMIFHeader);
                                             }
                                             else
                                                 row = ObjectColumns.Parse(l_line, mMIFHeader);
                                             mMIDData.mRows.Add(row);
                                             obj.Properties = mMIDData[mMIFData.Objects.Count - 1];
                                             continue;
                                         }
                                         else
                                             if (string.Compare(word, MIFWords.None, true) == 0)
                                             {
                                                 MIF.Primitives.MIFEmpty obj = new MIFEmpty();
                                                 mMIFData.Objects.Add(obj);
                                                 line = mif_reader.ReadLine();
                                                 string l_line = mid_reader.ReadLine();
                                                 ObjectColumns row;
                                                 if (l_line == null)
                                                 {
                                                     row = new ObjectColumns(mMIFHeader.mColumns.Length);
                                                     row.InitFromHeader(mMIFHeader);
                                                 }
                                                 else
                                                     row = ObjectColumns.Parse(l_line, mMIFHeader);
                                                 mMIDData.mRows.Add(row);
                                                 obj.Properties = mMIDData[mMIFData.Objects.Count - 1];
                                                 continue;
                                             }
                                             else
                                                 if (MIFData.IsDataObject(word))
                                                 {
                                                     mMIFData.GetStrObject(mif_reader, line, out line);
                                                     mid_reader.ReadLine();
                                                     continue;
                                                 }
                                             else
                                                return false;


                         }//end while
                     }//end using mid_reader
                 }//end using mif_reader
                 mMIFData.SetMapSize();
                 return true;
             }
             catch (Exception e)
             {
                 MessageBox.Show(e.Message);
                 return false;
             }
         }//end Read_new


         private bool Save(string aFileName)
         {
             CommandWriter commandWriter = new CommandWriter(aFileName, mMIFHeader.GetEncoding());
             using (commandWriter)
             {
                 if (!mMIFHeader.Save(commandWriter)) return false;
                 if (!mMIFData.Save(commandWriter)) return false;
                 if (!mMIDData.Save(aFileName.Substring(0, aFileName.IndexOf('.')) + ".MID", mMIFHeader)) return false;
                 if (!SaveGPS(aFileName.Substring(0, aFileName.IndexOf('.')) + ".GPS")) return false;
             }
             return true;
         }

         public bool SaveAs(string aFileName)
       {
           return Save(aFileName);
       }

         public bool Save()
         {
             return Save(mMifFileName);
         }
         
         public bool Draw(Bitmap aBitmap)
         {
             if (aBitmap == null) return false;
             return mMIFData.Draw(aBitmap);
         }

         public void RemoveAt(int index)
         {
             mMIDData.RemoveAt(index);
             mMIFData.RemoveAt(index);
         }
         public void Remove(BaseObject obj)
         {
             int idx = mMIFData.Objects.IndexOf(obj);
             mMIFData.Remove(obj);
             mMIDData.RemoveAt(idx);
         }

         public void Remove(ArrayList aBaseObjects)
         {
             if (aBaseObjects == null) return;
             for (int i = 0; i < aBaseObjects.Count; i++)
             {
                 BaseObject obj = aBaseObjects[i] as BaseObject;
                 int idx = mMIFData.Objects.IndexOf(obj);
                 mMIFData.RemoveAt(idx);
                 mMIDData.RemoveAt(idx);
             }
         }
        
         public int IndexOf(BaseObject obj)
            {
                return mMIFData.Objects.IndexOf(obj);
            }
         #endregion

         #region classes
         public class MIFHeader
         {
             #region fields
             public string mVersion;
             public string mCharSet;
             public string mDelimiter;
             public int[] mUnique;
             public int[] mIndex;
             public string mCoordSys;
             public string mTramsform;
             public MIFColumn[] mColumns;
             #endregion
             #region methods
             public static bool IsHeaderObject(string aString)
             {
                 if (aString == null) return false;
                 aString.Trim();
                 aString = aString.Substring(0, aString.IndexOf(' '));
                 if (
                     (String.Compare(aString, MIFWords.Version, true) != 0) &&
                     (String.Compare(aString, MIFWords.CharSet, true) != 0) &&
                     (String.Compare(aString, MIFWords.Delimiter, true) != 0) &&
                     (String.Compare(aString, MIFWords.Unique, true) != 0) &&
                     (String.Compare(aString, MIFWords.Index, true) != 0) &&
                     (String.Compare(aString, MIFWords.CoordSys, true) != 0) &&
                     (String.Compare(aString, MIFWords.Transform, true) != 0) &&
                     (String.Compare(aString, MIFWords.Columns, true) != 0)
                     )

                     return false;
                 return true;
             }
             public bool Save(CommandWriter aStreamWriter)
             {
                 try
                 {
                     aStreamWriter.WriteLine(MIFWords.Version + " " + mVersion);
                     aStreamWriter.WriteLine(MIFWords.CharSet + " \"" + mCharSet + "\"");
                     aStreamWriter.WriteLine(MIFWords.Delimiter + " \"" + mDelimiter + "\"");
                     if (mUnique.Length > 0)
                     {

                         aStreamWriter.Write(MIFWords.Unique + " ");
                         for (int i = 0; i < mUnique.Length; i++)
                         {
                             aStreamWriter.Write(mUnique[i].ToString());
                             if (i < mUnique.Length - 1) aStreamWriter.Write(",");

                         }
                         aStreamWriter.WriteLine();
                     }
                     if (mIndex.Length > 0)
                     {

                         aStreamWriter.Write(MIFWords.Index + " ");
                         for (int i = 0; i < mIndex.Length; i++)
                         {
                             aStreamWriter.Write(mIndex[i].ToString());
                             if (i < mIndex.Length - 1) aStreamWriter.Write(",");

                         }
                         aStreamWriter.WriteLine();
                     }
                     if (mCoordSys.Trim() != "")
                     {
                         aStreamWriter.WriteLine(string.Format("{0} {1}", MIFWords.CoordSys, mCoordSys));
                     }
                     if (mTramsform.Trim() != "")
                     {
                         aStreamWriter.WriteLine(string.Format("{0} {1}", MIFWords.Transform, mTramsform));
                     }
                     aStreamWriter.WriteLine(MIFWords.Columns + " " + mColumns.Length.ToString());
                     for (int i = 0; i < mColumns.Length; i++)
                     {
                         aStreamWriter.WriteLine("  " + mColumns[i].mName + " " + mColumns[i].mType);
                     }

                 }
                 catch
                 {
                     return false;
                 }
                 return true;
             }
             public Encoding GetEncoding()
             {
                 if (string.Compare(mCharSet, "WindowsCyrillic", true) == 0) return Encoding.GetEncoding(1251);
                 else return Encoding.Default;
             }
             public bool Read(CommandReader aCommandReader)
             {
                 try
                 {
                     string line = aCommandReader.ReadLine();
                     while (line != null)
                     {
                         line = line.Trim();
                         if (line.Length <= 0) { line = aCommandReader.ReadLine(); continue; }
                         string[] words = line.Split(new char[] { ' ', '\t', '\n', '\r' }, StringSplitOptions.RemoveEmptyEntries);
                         string word = words[0];
                         if (string.Compare(word, MIFWords.Version, true) == 0)
                         {
                             mVersion = words[1];
                         }
                         else
                             if (string.Compare(word, MIFWords.CharSet, true) == 0)
                             {
                                 mCharSet = words[1].Trim('"');
                             }
                             else
                                 if (string.Compare(word, MIFWords.Delimiter, true) == 0)
                                 {
                                     mDelimiter = words[1].Trim('"');
                                 }
                                 else
                                     if (string.Compare(word, MIFWords.Unique, true) == 0)
                                     {
                                         string[] tokens = words[1].Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries);
                                         mUnique = new int[tokens.Length];
                                         for (int i = 0; i < mUnique.Length; i++)
                                         {
                                             mUnique[i] = int.Parse(tokens[i]);
                                         }
                                     }
                                     else
                                         if (string.Compare(word, MIFWords.Index, true) == 0)
                                         {
                                             string[] tokens = words[1].Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries);
                                             mIndex = new int[tokens.Length];
                                             for (int i = 0; i < mIndex.Length; i++)
                                             {
                                                 mIndex[i] = int.Parse(tokens[i]);
                                             }
                                         }
                                         else
                                             if (string.Compare(word, MIFWords.CoordSys, true) == 0)
                                             {
                                                 mCoordSys = line.Replace(MIFWords.CoordSys, "").Trim();
                                                 string[] tokens = Strings.Split(mCoordSys, new char[] {',',' '});
                                                 //WGS84 datumn
                                                 if (string.Compare(tokens[0], "Earth") != 0) return false;
                                                 if (string.Compare(tokens[1], "Projection") != 0) return false;
                                                 if (string.Compare(tokens[2], "1")     != 0) return false;
                                                 if (string.Compare(tokens[3], "104")   != 0) return false;
                                             }
                                             else
                                                 if (string.Compare(word, MIFWords.Transform, true) == 0)
                                                 {
                                                     mTramsform = line.Replace(MIFWords.Transform, "").Trim();
                                                 }
                                                 else
                                                     if (string.Compare(word, MIFWords.Columns, true) == 0)
                                                     {
                                                         int num_cols = int.Parse(words[1]);
                                                         mColumns = new MIFColumn[num_cols];
                                                         for (int i = 0; i < num_cols; i++)
                                                         {
                                                             line = aCommandReader.ReadLine();
                                                             string[] tokens = line.Split(new char[] { ' ' }, StringSplitOptions.RemoveEmptyEntries);
                                                             MIFColumn col = new MIFColumn(tokens[0], tokens[1]);
                                                             mColumns[i] = col;
                                                         }
                                                     }
                                                     else
                                                         if (string.Compare(word, MIFWords.Data, true) == 0)
                                                         {
                                                             return true;
                                                         }

                         line = aCommandReader.ReadLine();
                     }
                     return false;
                 }
                 catch
                 {
                     return false;
                 }
             }
             public MIFHeader()
             {
                 mVersion = Defaults.Header_Version.ToString();
                 mDelimiter = Defaults.Header_Delimiter;
                 mCharSet = Defaults.Header_Charset;
                 mColumns = new MIFColumn[0];
                 mCoordSys = "";
                 mIndex = new int[0];
                 mUnique = new int[0];
                 mTramsform = "";
             }
             #endregion
         }

         public class MIFData
         {

             #region fields
             private ArrayList mObjects;
             public ArrayList Objects { get { return mObjects; } }

             private SizeDouble mMapSize;
             public SizeDouble MapSize { get { return mMapSize; } }

             public void SetMapSize()
             {
                 mMapSize = new SizeDouble(GetMaxX() - GetMinX(), GetMaxY() - GetMinY());
             }
             public BaseObject this[int index]
             {
                 get
                 {
                     if (!((0 <= index) && (index < mObjects.Count))) return null;
                     return mObjects[index] as BaseObject;
                 }
             }
             
             #endregion

             #region methods
             public void RemoveAt(int index)
             {
                 if ( (index<0) || (index>mObjects.Count-1) ) return;
                 mObjects.RemoveAt(index);
             }
             public void Remove(BaseObject obj)
             {
                 mObjects.Remove(obj);
             }
             public MIFData()
             {
                 mObjects = new ArrayList();
             }
             public double GetMinX()
             {
                 double l_min_x = double.MaxValue;
                 for (int i = 0; i < mObjects.Count; i++)
                 {
                     if ((mObjects[i] as BaseObject).MinX < l_min_x) l_min_x = (mObjects[i] as BaseObject).MinX;
                 }
                 return l_min_x;
             }
             public double GetMinY()
             {
                 double l_min_y = double.MaxValue;
                 for (int i = 0; i < mObjects.Count; i++)
                 {
                     if ((mObjects[i] as BaseObject).MinY < l_min_y) l_min_y = (mObjects[i] as BaseObject).MinY;
                 }
                 return l_min_y;
             }
             public double GetMaxX()
             {
                 double l_max_x = double.MinValue;
                 for (int i = 0; i < mObjects.Count; i++)
                 {
                     if ((mObjects[i] as BaseObject).MaxX > l_max_x) l_max_x = (mObjects[i] as BaseObject).MaxX;
                 }
                 return l_max_x;
             }
             public double GetMaxY()
             {
                 double l_max_y = double.MinValue;
                 for (int i = 0; i < mObjects.Count; i++)
                 {
                     if ((mObjects[i] as BaseObject).MaxY > l_max_y) l_max_y = (mObjects[i] as BaseObject).MaxY;
                 }
                 return l_max_y;
             }
             public void Clear()
             {
                 mObjects.Clear();
                 mObjects.TrimToSize();

             }
             public static bool IsDataObject(string aString)
             {
                 if (aString == null) return false;
                 if (aString.Length <= 0) return false;
                 aString.Trim();
                 aString = aString.Split(new char[] { ' ' })[0];
                 if (
                     (String.Compare(aString, MIFWords.Point, true) != 0) &&
                     (String.Compare(aString, MIFWords.Line, true) != 0) &&
                     (String.Compare(aString, MIFWords.Pline, true) != 0) &&
                     (String.Compare(aString, MIFWords.Region, true) != 0) &&
                     (String.Compare(aString, MIFWords.Arc, true) != 0) &&
                     (String.Compare(aString, MIFWords.Text, true) != 0) &&
                     (String.Compare(aString, MIFWords.Rect, true) != 0) &&
                     (String.Compare(aString, MIFWords.RoundRect, true) != 0) &&
                     (String.Compare(aString, MIFWords.Ellipse, true) != 0) &&
                     (String.Compare(aString, MIFWords.MultiPoint, true) != 0) &&
                     (String.Compare(aString, MIFWords.Collection, true) != 0) &&
                     (String.Compare(aString, MIFWords.None, true) != 0)
                     )

                     return false;

                 return true;
             }
             private string[] ObjectsToStrings()
             {
                 string[] result = new string[mObjects.Count];
                 int i = 0;
                 foreach (BaseObject o in mObjects)
                 {
                     result[i] = o.ToString();
                     i++;
                 }
                 return result;
             }
             public string GetStrObject(CommandReader aCommandReader, string aLine, out string line)
             {
                 StringBuilder sb = new StringBuilder(aLine);
                 aLine = aCommandReader.ReadLine();
                 while ((aLine != null) && (!IsDataObject(aLine)))
                 {
                     sb.Append("\r\n");
                     sb.Append(aLine);
                     aLine = aCommandReader.ReadLine();
                 }
                 line = aLine;
                 return sb.ToString();
             }
             private string GetStrObjectCollection(CommandReader aCommandReader, string aLine, out string line)
             {
                 StringBuilder sb = new StringBuilder(aLine);
                 string[] tokens = aLine.Split(new char[] { ' ' });
                 int num_parts = 3; if (tokens.Length == 2) num_parts = int.Parse(tokens[1]);
                 aLine = aCommandReader.ReadLine();
                 while ((aLine != null) && (num_parts > 0))
                 {
                     sb.Append("\n\r");
                     sb.Append(aLine);
                     aLine = aCommandReader.ReadLine();
                     if (IsDataObject(aLine)) num_parts--;
                 }
                 line = aLine;
                 return sb.ToString();
             }
             public bool Read(CommandReader aCommandReader, MIDData aMIDData)
             {
                 try
                 {
                     Clear();
                     string line = aCommandReader.ReadLine();
                     while (line != null)
                     {
                         line = line.Trim();
                         if (line.Length <= 0) { line = aCommandReader.ReadLine(); continue; }
                         string[] words = line.Split(new char[] { ' ', '\t', '\n', '\r' });
                         string word = words[0];
                         if (string.Compare(word, MIFWords.Point, true) == 0)
                         {
                             MIF.Primitives.MIFPoint obj = MIF.Primitives.MIFPoint.Parse(GetStrObject(aCommandReader, line, out line));
                             mObjects.Add(obj);
                             obj.Properties = aMIDData[mObjects.Count - 1];
                             continue;

                         }
                         else
                             if (string.Compare(word, MIFWords.Line, true) == 0)
                             {
                                 MIF.Primitives.MIFLine obj = MIF.Primitives.MIFLine.Parse(GetStrObject(aCommandReader, line, out line));
                                 mObjects.Add(obj);
                                 obj.Properties = aMIDData[mObjects.Count - 1];
                                 continue;
                             }

                             else
                                 if (string.Compare(word, MIFWords.Pline, true) == 0)
                                 {
                                     MIF.Primitives.MIFPolyLine obj = MIF.Primitives.MIFPolyLine.Parse(GetStrObject(aCommandReader, line, out line));
                                     mObjects.Add(obj);
                                     obj.Properties = aMIDData[mObjects.Count - 1];
                                     continue;
                                 }

                                 else
                                     if (string.Compare(word, MIFWords.Region, true) == 0)
                                     {

                                         MIF.Primitives.MIFRegion obj = MIF.Primitives.MIFRegion.Parse(GetStrObject(aCommandReader, line, out line));
                                         mObjects.Add(obj);
                                         obj.Properties = aMIDData[mObjects.Count - 1];
                                         continue;
                                     }
                                     #region TODO other primitives
                                     /*
                                                      else
                                                   if (string.Compare(word, MIFWords.Arc,true) == 0)
                                                   {

                                                       MIF.Primitives.MIFArc obj = MIF.Primitives.MIFArc.Parse(GetStrObject(line, out line));
                                                       mObjects.Add(obj);
                                                       continue;
                                                   }
                     
                                               else
                                                       if (string.Compare(word, MIFWords.Text, true) == 0)
                                                 {
                                                     MIF.Primitives.MIFText obj = MIF.Primitives.MIFText.Parse(GetStrObject(line, out line));
                                                     mObjects.Add(obj);
                                                     continue;
                                                 }
                 
                                            else
                                              if (string.Compare(word, MIFWords.Rect, true) == 0)
                                              {
                                                  MIF.Primitives.MIFRectangle obj = MIF.Primitives.MIFRectangle.Parse(GetStrObject(line, out line));
                                                  mObjects.Add(obj);
                                                  continue;
                                              }
                  
                                               else
                                                  if (string.Compare(word, MIFWords.RoundRect, true) == 0)
                                                 {
                                                     MIF.Primitives.MIFRndRectangle obj = MIF.Primitives.MIFRndRectangle.Parse(GetStrObject(line, out line));
                                                         mObjects.Add(obj);
                                                         continue;
                                                 }
                      
                                                 else
                                                     if (string.Compare(word, MIFWords.Ellipse, true) == 0)
                                                   {
                                                       MIF.Primitives.MIFEllipse l_obj = MIF.Primitives.MIFEllipse.Parse(GetStrObject(line, out line));
                                                           mObjects.Add(l_obj);
                                                           continue;
                                                   }
                                                else
                                                       if (string.Compare(word, MIFWords.MultiPoint, true) == 0)
                                                      {
                                                          MIF.Primitives.MIFMultipoint l_obj =  MIF.Primitives.MIFMultipoint.Parse(GetStrObject(line, out line));
                                                          mObjects.Add(l_obj);
                                                          continue;
                                                      }
                          
                                                         else
                                                           if (string.Compare(word, MIFWords.Collection, true) == 0)
                                                         {
                                                                 MIF.Primitives.MIFCollection l_obj = MIF.Primitives.MIFCollection.Parse(GetStrObjectCollection(line, out line));
                                                                 mObjects.Add(l_obj);
                                                                 continue;
                                                         } */
                                     #endregion
                                     else
                                         if (string.Compare(word, MIFWords.None, true) == 0)
                                         {
                                             mObjects.Add(new MIFEmpty());
                                             (mObjects[mObjects.Count - 1] as BaseObject).Properties = aMIDData[mObjects.Count - 1];
                                             line = aCommandReader.ReadLine();
                                         }
                                         else
                                             return false;

                     }
                     mMapSize = new SizeDouble(GetMaxX() - GetMinX(), GetMaxY() - GetMinY());
                     return true;
                 }
                 catch (ApplicationException e)
                 {
                     MessageBox.Show(e.Message);
                     return false;
                 }
             }//end read
             public bool Save(CommandWriter aStreamWriter)
             {
                 aStreamWriter.WriteLine(MIFWords.Data);
                 aStreamWriter.WriteLine("");
                 foreach (BaseObject o in mObjects)
                 {
                     if (!o.Save(aStreamWriter)) return false;
                 }

                 return true;
             }
             public bool Draw(Bitmap aBitmap)
             {

                 using (Graphics gr = Graphics.FromImage(aBitmap))
                 {
                     gr.Clear(Color.White);
                     for (int i = 0; i < Objects.Count; i++)
                         if (!(Objects[i] as BaseObject).Draw(aBitmap)) return false;
                     return true;
                 }
             }
             #endregion

         }

         public class MIDData
         {
             public ObjectColumns_S mRows;
             public ObjectColumns this[int index]
             {
                 get
                 {
                     if (index >= mRows.Count) return null;
                     return mRows[index] as ObjectColumns;
                 }
             }
             public MIDData()
             {
                 mRows = new ObjectColumns_S();
             }

             public void RemoveAt(int index)
             {
                 if ((index < 0) || (index > mRows.Count - 1)) return;
                 mRows.RemoveAt(index);
             }
             public void Remove(ObjectColumns obj)
             {
                 mRows.Remove(obj);
             }
             public void Clear()
             {
                 mRows.Clear();
             }

            

             public bool Save(string aFileName, MIFHeader aHeader)
             {
                 using (CommandWriter streamWriter = new CommandWriter(aFileName, aHeader.GetEncoding()))
                 {
                     try
                     {
                         foreach (ObjectColumns row in mRows)
                         {
                             streamWriter.WriteLine(row.ConvertToString(aHeader.mDelimiter.ToCharArray()[0]));
                         }
                         return true;
                     }
                     catch
                     {
                         return false;
                     }
                 }
             }


         }
         #endregion
     }

    


     



  

}
