using System;
using System.IO;
using System.Collections;
using System.Text;
using System.Windows.Forms;
using System.Drawing;
using MyUtils;
using System.Drawing.Imaging;
using System.Threading;
using MyNMEA;




/// <summary>
/// Summary description for Class1
/// </summary>
/// 

namespace MIF.Primitives
{
    public enum CoordSys {WGS84,UTM}

    public class GPS
    {
        private Rectangle m_PosRect;
        private string m_GPSRawMsg;

        public string GPSRawMsg
        {
            get { return m_GPSRawMsg; }
        }
        public  PointDouble Coordinates;
        public  Point PixCoordinates;
        public  GPS()
        {
            Coordinates     = new PointDouble();
            PixCoordinates  = new Point();
            m_PosRect       = new Rectangle();
        }
        public void DrawPosition(Graphics gr)
        {
            
            m_PosRect.Width = m_PosRect.Height = 30;
            m_PosRect.X = PixCoordinates.X - m_PosRect.Width/2;
            m_PosRect.Y = PixCoordinates.Y - m_PosRect.Height/2;
            Pen pen = new Pen(Color.Red,3);
            gr.DrawEllipse(pen, m_PosRect);
            gr.DrawLine(Pens.Black, new Point(m_PosRect.X + 10, m_PosRect.Y + m_PosRect.Height / 2), new Point(m_PosRect.X + m_PosRect.Width - 10, m_PosRect.Y + m_PosRect.Height / 2));
            gr.DrawLine(Pens.Black, new Point(m_PosRect.X + m_PosRect.Width/2, m_PosRect.Y + 10), new Point(m_PosRect.X + m_PosRect.Width/2, m_PosRect.Y + m_PosRect.Height -10 ));
            pen.Dispose();
        }
      
        public void ReadData()
        {
            try
            {
                using (System.IO.StreamReader sr = new System.IO.StreamReader(@"D:\Niki\Projects\MIFFile\test_data\NMEA.txt"))
                {
                    string msg = sr.ReadLine();
                    int cnt_msg = 0; 
                    while ((msg != null) && (msg.Trim() != ""))
                    {
                        MyNMEA.NMEA en_nmea = MyNMEA.NMEA.None;
                        if (!MyNMEA.NMEA_Msg.IsCorrectNMEA(msg, out en_nmea)) { msg = sr.ReadLine(); continue; }
                        MyNMEA.NMEA_Msg msg_nmea = null;
                        switch (en_nmea)
                        {
                            case MyNMEA.NMEA.GGA: msg_nmea = MyNMEA.GGA.Parse(msg);
                                if (!msg_nmea.IsValid()) { msg = sr.ReadLine(); continue; }
                                Coordinates.X = (msg_nmea as MyNMEA.GGA).LongitudeDeg;
                                Coordinates.Y = (msg_nmea as MyNMEA.GGA).LatitudeDeg;                 
                            break;
                            case MyNMEA.NMEA.GLL: msg_nmea = MyNMEA.GLL.Parse(msg);
                                if (!msg_nmea.IsValid()) { msg = sr.ReadLine(); continue; }
                                Coordinates.X = (msg_nmea as MyNMEA.GLL).LongitudeDeg;
                                Coordinates.Y = (msg_nmea as MyNMEA.GLL).LatitudeDeg; 
                            break;
                            case MyNMEA.NMEA.GSA: msg_nmea = MyNMEA.GSA.Parse(msg); break;
                            case MyNMEA.NMEA.GSV: msg_nmea = MyNMEA.GSV.Parse(msg); break;
                            case MyNMEA.NMEA.RMC: msg_nmea = MyNMEA.RMC.Parse(msg);
                                if (!msg_nmea.IsValid()) { msg = sr.ReadLine(); continue; }
                                Coordinates.X = (msg_nmea as MyNMEA.RMC).LongitudeDeg;
                                Coordinates.Y = (msg_nmea as MyNMEA.RMC).LatitudeDeg; 
                            break;
                            case MyNMEA.NMEA.VTG: msg_nmea = MyNMEA.VTG.Parse(msg); break;
                            default: return ;
                        }
                        MapTrans.BasePoint = Coordinates;
                        int x, y;
                        MapTrans.GetPixel(Coordinates.X, Coordinates.Y, out x, out y);
                        PixCoordinates.X = x;
                        PixCoordinates.Y = y;
                        cnt_msg++;
                        if (cnt_msg == 6)
                        {
                            cnt_msg = 0;
                            Thread.Sleep(1000);
                        }
                        msg = sr.ReadLine();
                    }
                }
                return ;
            }
            catch
            {
                return;
            }
        }
        public void ReadFromParser(CNMEAParser aParser)
        {
            Coordinates.X = aParser.m_dRMCLongitude;
            Coordinates.Y = aParser.m_dRMCLatitude;
            m_GPSRawMsg   = aParser.RawRMC;
            MapTrans.BasePoint = Coordinates;
            int x, y;
            MapTrans.GetPixel(Coordinates.X, Coordinates.Y, out x, out y);
            PixCoordinates.X = x;
            PixCoordinates.Y = y;
        }

    }
    
    
    
    public class MIFColumn
    {
        public string mName;
        public string mType;
        public MIFColumn(string aName, string aType)
        {
            mName = aName;
            mType = aType;
        }

        public static MIFColumn Parse(string aLine)
        {

            aLine = aLine.Trim();
            string[] l_fields = aLine.Split(' ');
            MIFColumn result = new MIFColumn(l_fields[0], l_fields[1]);
            return result;
        }
        public override string ToString()
        {
            return (mName + " ( " + mType + " ) ");
        }

        public override int GetHashCode()
        {
            return mName.GetHashCode();
        }
    }
    
    public class ObjectColumn
    {
        public string Name="";
        public string Value="";
        public string Type="";
        public int Width;
        public ObjectColumn (string aName, string aValue, string aType)
        {
            if (aName!=null)
                Name = aName;
            if (aValue!=null)
                Value = aValue;
            if (aType != null)
                Type = aType;
        }

        public bool GetMIFTypeAndWidth(out string aMIFType,out int aWidth,out int aDecimals)
        {
            aMIFType = ""; aWidth = 0; aDecimals = 0;
            if (Type == null) return false;
            string[] tokens = Strings.Split(Type, new char[] {' ','\t','(',')',',','.'});
            if (tokens.Length <= 0) return false;
            aMIFType = tokens[0];
            if (tokens.Length > 1)
                aWidth = int.Parse(tokens[1]);
            if (tokens.Length > 2)
                aDecimals = int.Parse(tokens[2]);
            return true;
        }

        

        public string MakeRow(string BorderStr)
        {
            StringBuilder sb = new StringBuilder();
            sb.Append(Name.PadRight(10)); sb.Append(BorderStr);
            sb.Append(Value.PadRight(10));
            return sb.ToString();
        }
        public static string MakeRowLabel(string aBorderStr)
        {
            StringBuilder sb = new StringBuilder();
            sb.Append("".PadRight(10)); sb.Append(aBorderStr);
            sb.Append("".PadRight(10));
            return sb.ToString();
        }
    }

    public class ObjectColumns : ArrayList
    {
        public int Add(ObjectColumn value)
        {
            //if (!(value is ObjectColumn)) throw new ApplicationException("Only ObjectColumn value can be added!");
            return base.Add(value);
        }
        public static ObjectColumns Parse(string aString,MIFFile.MIFHeader aHeader)
        {
            
            string[] columns = aString.Split(aHeader.mDelimiter.ToCharArray());
            string[] adjustedcols = new string[aHeader.mColumns.Length];
            for (int i = 0; i < adjustedcols.Length; i++) adjustedcols[i] = "";
                if (columns != null)
                {
                    int j = 0;
                    for (int i = 0; i < columns.Length; i++)
                    {
                        adjustedcols[j] += columns[i];
                        if (adjustedcols[j].Trim().Length <= 0) { j++; continue; }
                        if ((adjustedcols[j].Trim()[0] == '"') && (adjustedcols[j].Trim()[adjustedcols[j].Trim().Length - 1] != '"')) continue;
                        j++;
                    }
                }
            
            ObjectColumns result = new ObjectColumns(adjustedcols.Length);
            for (int i = 0; i < adjustedcols.Length; i++)
            {
                adjustedcols[i] = adjustedcols[i].TrimStart(new char[] { '"' });
                adjustedcols[i] = adjustedcols[i].TrimEnd(new char[] { '"' });
                ObjectColumn col = new ObjectColumn(aHeader.mColumns[i].mName,adjustedcols[i],aHeader.mColumns[i].mType);
                result.Add(col);
            }
            return result;
        }
 
        public string ConvertToString(char aDelimiter)
        {
            StringBuilder sb = new StringBuilder(aDelimiter);
            for (int i = 0; i < Count; i++)
            {
                bool HasKV = ((this[i] as ObjectColumn).Type.Trim().StartsWith(MIFTypes.Character,StringComparison.InvariantCultureIgnoreCase));
                if (HasKV) sb.Append('"');
                sb.Append((this[i] as ObjectColumn).Value);
                if (HasKV) sb.Append('"');
                if (i<Count-1) sb.Append(aDelimiter);
            }
            return sb.ToString();
        }
      
        public ObjectColumns(int aSize):base(aSize)
        {
           

        }
        public void InitFromHeader(MIFFile.MIFHeader aMIFHeader)
        {
            for (int i = 0; i < aMIFHeader.mColumns.Length; i++)
            {
                Add(new ObjectColumn(aMIFHeader.mColumns[i].mName, "\"\"", aMIFHeader.mColumns[i].mType));
            }
        }
    }

    public class ObjectColumns_S : ArrayList
    {
        public int Add(ObjectColumns value)
        {
            return base.Add(value);
        }
    }
  


    public abstract class BaseObject
    {
        protected bool mIsSelected=false;
        protected double mMinX=0;
        protected double mMinY=0;
        protected double mMaxX=0;
        protected double mMaxY=0;
        protected ArrayList m_GPSRawStrings = new ArrayList();
        public ArrayList GPSRawStrings
        {
            get { return m_GPSRawStrings; }
            set { m_GPSRawStrings = value; }
        }
        public double MinX { get { return mMinX; } }
        public double MinY { get { return mMinY; } }
        public double MaxX { get { return mMaxX; } }
        public double MaxY { get { return mMaxY; } }
        public abstract bool IsSelected { get;set; }
        public bool Save(CommandWriter aStreamWriter)
        {
            try
            {
                string[] rows = this.ToString().Split(new char[] { '\n', '\r' }, StringSplitOptions.RemoveEmptyEntries);
                for (int i = 0; i < rows.Length; i++)
                {
                    aStreamWriter.WriteLine(rows[i]);
                }
            }
            catch
            {
                return false;
            }
          
            return true;
        }
        public override string ToString()
        {
            return MIFWords.None;
        }
        public abstract bool Draw(Bitmap aBitmap);
        public abstract bool DoSelect(double x, double y,double select_distance);
        public ObjectColumns Properties;//pointer
    }
   
    public class MIFPoint:BaseObject
    {
        
        public  MIFSymbol mSymbol;
        private double mX;
        private double mY;
        
        
 
        public double X
        {
            get {return mX ;}
        }
        public double Y
        {
            get { return mY; }
        }
        public MIFPoint(double aX, double aY,MIFSymbol aSymbol)
        {
            mX = aX;
            mY = aY;
            if (aSymbol == null) mSymbol = new MIFSymbol();
            else
            mSymbol = new MIFSymbol(aSymbol.Shape, aSymbol.Color, aSymbol.Size);
            mMinX=mMaxX=mX;
            mMinY=mMaxY=mY;
        }
        static public MIFPoint Parse(string aString)
        {
            try
            {
                MIFPoint result = null;
                string[] tokens = Strings.Split(aString, new char[] { ' ', '\t', ',', '(', ')', '\n', '\r' });
                if (string.Compare(tokens[0], MIFWords.Point, true) != 0) throw new ApplicationException("Error MIFPoint.Parse()");

                if ((tokens.Length == 7) && (string.Compare(tokens[3], MIFWords.Symbol, true) == 0))
                {
                    result = new MIFPoint(Double.Parse(tokens[1]), Double.Parse(tokens[2]), new MIFSymbol(int.Parse(tokens[4]), int.Parse(tokens[5]), int.Parse(tokens[6])));

                }
                else
                {
                    result = new MIFPoint(Double.Parse(tokens[1]), Double.Parse(tokens[2]),null);
                }
                
                return result;
            }
            catch 
            {
                throw new ApplicationException("Invalid Point string"+"\r\n"+aString+"\r\n");
            }
            
        }
        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();
            sb.Append(string.Format("{0} {1} {2}\r\n",MIFWords.Point,mX,mY));
            sb.Append(string.Format("     {0}",mSymbol));
            return sb.ToString();
        }
        public override bool Draw(Bitmap aBitmap)
        {
            try
            {
                Graphics gr = Graphics.FromImage(aBitmap);
                int x, y;
                MapTrans.GetPixel(X, Y, out x, out y);
                mSymbol.Draw(x,y,gr,mIsSelected);
                gr.Dispose();
                return true;
            }
            catch
            {
                return false;
            }
        }
        public override bool DoSelect(double x, double y, double select_distance)
        {
            Point point_pixels = MapTrans.GetPixel(new PointDouble(mX,mY));
            point_pixels.X += mSymbol.Size / 2;
            point_pixels.Y += mSymbol.Size / 2;
            PointDouble pnt = MapTrans.GetDegree(point_pixels);
            return (Geometry.DistFromDegPointToDegPointInPixels(x , y , pnt.X, pnt.Y, MapTrans.mPixelsPerDegree) <= mSymbol.Size*Math.Sqrt(2)/2);
        }
        public override bool IsSelected
        {
            get { return mIsSelected;  }
            set { mIsSelected = value; }
        }
    }

    public class MIFLine:BaseObject
    {
        public const int MinWidth = 1;
        public const int MaxWidth = 50;
        private MIFPen  mMIFPen;
        public MIFPen MIFPen
        {
            get { return mMIFPen; }
        }
        private MIFNode mNode1;
        private MIFNode mNode2;


        public static void PutToComboLineWidth(ComboBox.ObjectCollection aItems)
        {
            aItems.Clear();
            for (int i = MinWidth; i < MaxWidth + 1; i++)
            {
                Strings.ItemsAdd(aItems,string.Format("Line width {0} ",i),i.ToString());
            }
        }

        public override bool IsSelected
        {
            get { return mIsSelected; }
            set { mIsSelected =value; }
        }

        public override bool DoSelect(double x, double y, double select_distance)
        {
            if ((Geometry.DistDegLineToDegPointInPixels(mNode1.X, mNode1.Y, mNode2.X, mNode2.Y, x, y,MapTrans.mPixelsPerDegree) <= select_distance)) {
                
                
                return true;
            }
            return false;
            
        }

        public override bool Draw(Bitmap aBitmap)
        {
            Graphics gr=null;
            Pen pen=null;
            
            try
            {
                gr = Graphics.FromImage(aBitmap);
                if (mIsSelected)
                    pen = new Pen(Defaults.SelectColor,mMIFPen.Width-1 );
                else
                    pen = mMIFPen.NewPen();
                PointF p1 = new PointF(0, 0);
                PointF p2 = new PointF(0, 0);
                int x, y, x1, y1;
                MapTrans.GetPixel(mNode1.X, mNode1.Y, out x, out y);
                MapTrans.GetPixel(mNode2.X, mNode2.Y, out x1, out y1);
                p1.X = x; p1.Y = y;
                p2.X = x1; p2.Y = y1;
                gr.DrawLine(pen, p1, p2);
                return true;
            }
            catch
            {
                return false;
            }
            finally
            {
                pen.Dispose();
                gr.Dispose();
            }
        }

        public static MIFLine Parse(string aString) 
        {
            try
            {
                MIFLine result = null;
                string[] tokens = Strings.Split(aString,new char[] { ' ', '\t', ',', '(', ')', '\n', '\r' });
                if (string.Compare(tokens[0], MIFWords.Line, true) != 0) throw new ApplicationException("Error MIFLine.Parse()");

                if ((tokens.Length == 9) && (string.Compare(tokens[5], MIFWords.Pen) == 0))
                {
                    result = new MIFLine(new MIFNode(Double.Parse(tokens[1]), Double.Parse(tokens[2]),true), new MIFNode(Double.Parse(tokens[3]), Double.Parse(tokens[4]),true), new MIFPen(int.Parse(tokens[6]), int.Parse(tokens[7]), int.Parse(tokens[8])));
                }
                else
                {
                    result = new MIFLine(new MIFNode(Double.Parse(tokens[1]), Double.Parse(tokens[2]),true), new MIFNode(Double.Parse(tokens[3]), Double.Parse(tokens[4]),true),null);
                }
                return result;
            }
            catch 
            {
                throw new ApplicationException("Invalid Line string " + "\r\n" + aString);
            }
           
        }
        
        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();
            sb.Append(string.Format("{0} {1} {2}\r\n",MIFWords.Line,mNode1,mNode2));
            sb.Append(string.Format("\t{0}",mMIFPen));
            return sb.ToString();
        }

        public MIFLine(MIFNode aNode1, MIFNode aNode2, MIFPen aPen)
        {
            mNode1 = aNode1;
            mNode2 = aNode2;
            if (aPen == null) mMIFPen = new MIFPen();else
            mMIFPen   =  new MIFPen(aPen);
            mMinX = Math.Min(mNode1.X, mNode2.X);
            mMinY = Math.Min(mNode1.Y, mNode2.Y);
            mMaxX = Math.Max(mNode1.X, mNode2.X);
            mMaxY = Math.Max(mNode2.Y, mNode2.Y);  
        }
        
        

    }

    public class MIFPolyLine : BaseObject
    {
        private bool mIsSmooth;
        private  MIFPen mMIFPen;
        public MIFPen MIFPen
        {
            get { return mMIFPen; }
        }
        public void SetMIFPenWidth(int aWidth)
        {
            mMIFPen.Width = aWidth;
            for (int i = 0; i < mSections.Count; i++)
            {
                (mSections[i] as MIFSection).MIFPen.Width = mMIFPen.Width;
            }
        }
        public void SetMIFPenColor(int aRGB)
        {
            mMIFPen.MIFColor = aRGB;
            for (int i = 0; i < mSections.Count; i++)
            {
                (mSections[i] as MIFSection).MIFPen.MIFColor = mMIFPen.MIFColor;
            }
        }
        public ArrayList mSections;


        public void RemoveNodeAt(int aSection, int Idx)
        {
            (mSections[aSection] as MIFSection).RemoveAt(Idx);

            if (((mSections[aSection] as MIFSection).Nodes.Count <= 0)) 
                mSections.RemoveAt(aSection);
            
        }

        private double GetMinX()
        {
            double l_min_X = double.MaxValue;
            for (int i = 0; i < mSections.Count; i++)
            {
                if ((mSections[i] as MIFSection).MinX < l_min_X) l_min_X = (mSections[i] as MIFSection).MinX;
            }
            return l_min_X;
        }

        private double GetMinY()
        {
            double l_min_Y = double.MaxValue;
            for (int i = 0; i < mSections.Count; i++)
            {
                if ((mSections[i] as MIFSection).MinY < l_min_Y) l_min_Y = (mSections[i] as MIFSection).MinY;
            }
            return l_min_Y;
        }

        private double GetMaxX()
        {
            double l_max_X = double.MinValue;
            for (int i = 0; i < mSections.Count; i++)
            {
                if ((mSections[i] as MIFSection).MaxX > l_max_X) l_max_X = (mSections[i] as MIFSection).MaxX;
            }
            return l_max_X;
        }

        private double GetMaxY()
        {
            double l_max_Y = double.MinValue;
            for (int i = 0; i < mSections.Count; i++)
            {
                if ((mSections[i] as MIFSection).MaxY > l_max_Y) l_max_Y = (mSections[i] as MIFSection).MaxY;
            }
            return l_max_Y;
        }





        public void Clear()
        {
            mSections.Clear();
        }
        
        public void AddSection(MIFSection aSection)
        {
            mSections.Add(aSection);
            mMinX = Math.Min(mMinX,aSection.MinX);
            mMinY = Math.Min(mMinY,aSection.MinY);
            mMaxX = Math.Max(mMaxX,aSection.MaxX);
            mMaxY = Math.Max(mMaxY,aSection.MaxY);
        }
        


        public override bool IsSelected
        {
            get { return mIsSelected; }
            set {
                mIsSelected = value; 
                for (int i = 0; i < mSections.Count; i++)
                    (mSections[i] as MIFSection).IsSelected = mIsSelected;
                }
        }

        public override bool DoSelect(double x, double y, double select_distance)
        {
            for (int i = 0; i < mSections.Count; i++)
            {
                if ((mSections[i] as BaseObject).DoSelect(x, y, select_distance)) return true;
            }
            return false;
        }
        public override bool Draw(Bitmap aBitmap)
        {
            for (int i = 0; i < mSections.Count; i++)
            {
                if (!(mSections[i] as BaseObject).Draw(aBitmap)) return false;
            }
            return true;
        }

        public MIFPolyLine(bool IsSmooth, MIFPen aPen, ArrayList aSections)
        {
            mSections = aSections;
            mIsSmooth = IsSmooth;
            if (aPen == null) mMIFPen = new MIFPen();
            else
                mMIFPen = new MIFPen(aPen);
            mMinX = GetMinX();
            mMinY = GetMinY();
            mMaxX = GetMaxX();
            mMaxY = GetMaxY();
            
        }

        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();
            sb.Append(MIFWords.Pline);
            if (mSections.Count > 1)
                sb.Append(string.Format(" {0} {1}\r\n",MIFWords.Multiple,mSections.Count));
            for (int i = 0; i < mSections.Count; i++)
            {
                sb.Append(mSections[i]);
            }
            sb.Append(mMIFPen);

            return sb.ToString();
        }
        public static MIFPolyLine Parse(string aString)
        {
            try
            {
                string[] tokens = Strings.Split(aString, new char[] { ' ', '\t', ',', '(', ')', '\n', '\r' });
                if (string.Compare(tokens[0], MIFWords.Pline, true) != 0) throw new ApplicationException("Invalid PLine string");
                int sections_count = 1;
                int current_token = 0;
                if (string.Compare(tokens[1], MIFWords.Multiple) == 0) { sections_count = int.Parse(tokens[2]); current_token = 1; };
                ArrayList sections = new ArrayList(sections_count);
                for (int i = 0; i < sections_count; i++)
                {
                    int nodes_count = int.Parse(tokens[++current_token]);
                    ArrayList nodes = new ArrayList(nodes_count);
                    for (int j = 0; j < nodes_count; j++)
                    {
                        MIFNode node = new MIFNode(double.Parse(tokens[++current_token]), double.Parse(tokens[++current_token]),true);

                        nodes.Add(node);
                    }
                    MIFSection section = new MIFSection(nodes,SectionType.PolyLine,new MIFPen(),new MIFBrush());
                    sections.Add(section);
                }

                MIFPen l_Pen = new MIFPen();
                bool issmooth = false;

                current_token++;
                if  (current_token < tokens.Length)
                {
                    if (string.Compare(tokens[current_token], MIFWords.Pen, true) == 0)
                    {
                        l_Pen = new MIFPen(int.Parse(tokens[++current_token]), int.Parse(tokens[++current_token]), int.Parse(tokens[++current_token]));

                    }
                    current_token++;
                }
                if (current_token < tokens.Length)
                {
                    if (string.Compare(tokens[current_token], MIFWords.Smooth) == 0)
                    {
                        issmooth = true;
                    }
                }
                for (int i = 0; i < sections.Count; i++)
                {
                    (sections[i] as MIFSection).MIFPen = l_Pen;
                }

                    return new MIFPolyLine(issmooth, l_Pen, sections);

            }

            catch
            {
                throw new ApplicationException("Invalie PLine string" + aString);
            }
        }
    }
    
    public class MIFRegion:BaseObject
    {
        
        public ArrayList   mSections;
        public  MIFPen      mMIFPen;
        public  MIFPen      MIFPen
        {
            get { return mMIFPen;}
        }
        private MIFBrush    mMIFBrush;
        public  MIFBrush MIFBrush
        {
            get { return mMIFBrush; }
        }
        private MIFNode     mCenter;


        public void SetMIFPenWidth(int aWidth)
        {
            mMIFPen.Width = aWidth;
            for (int i = 0; i < mSections.Count; i++)
            {
                (mSections[i] as MIFSection).MIFPen.Width = mMIFPen.Width;
            }
        }
        public void SetMIFPenColor(int aRGB)
        {
            mMIFPen.MIFColor = aRGB;
            for (int i = 0; i < mSections.Count; i++)
            {
                (mSections[i] as MIFSection).MIFPen.MIFColor = mMIFPen.MIFColor;
            }
        }
        public void SetMIFBrushColor(int aRGB)
        {
            mMIFBrush.MIFBackColor = aRGB;
            for (int i = 0; i < mSections.Count; i++)
            {
                (mSections[i] as MIFSection).MIFBrush.MIFBackColor = mMIFBrush.MIFBackColor;
            }
        }

        public void RemoveNodeAt(int aSection, int Idx)
        {
            (mSections[aSection] as MIFSection).RemoveAt(Idx);

            if (((mSections[aSection] as MIFSection).Nodes.Count <= 0))
                mSections.RemoveAt(aSection);
        }


        public void AddSection(MIFSection aSection)
        {
            mSections.Add(aSection);
            mMinX = Math.Min(mMinX, aSection.MinX);
            mMinY = Math.Min(mMinY, aSection.MinY);
            mMaxX = Math.Max(mMaxX, aSection.MaxX);
            mMaxY = Math.Max(mMaxY, aSection.MaxY);
        }

        public override int GetHashCode()
        {
            return mSections.GetHashCode();
        }
        
        public override bool Equals(object obj)
        {
            if (obj == null || GetType() != obj.GetType()) 
                return false;

            return mSections.Equals((obj as MIFRegion).mSections );
        }
        
        public override bool IsSelected
        {
            get { return mIsSelected; }
            set 
            {
                mIsSelected = value;
                for (int i = 0; i < mSections.Count; i++)
                {
                    (mSections[i] as MIFSection).IsSelected = mIsSelected;
                }
            }
        }

        public override bool DoSelect(double x, double y, double select_distance)
        {
            for (int i = 0; i < mSections.Count; i++)
            {
                if ((mSections[i] as MIFSection).DoSelect(x, y,select_distance)) 
                {
                    
                    return true;
                }
            }
            return false;
        }
            


        private double GetMinX()
        {
            double l_min_X = double.MaxValue;
            for (int i = 0; i < mSections.Count; i++)
            {
                if ((mSections[i] as MIFSection).MinX < l_min_X) l_min_X = (mSections[i] as MIFSection).MinX;
            }
            return l_min_X;
        }

        private double GetMinY()
        {
            double l_min_Y = double.MaxValue;
            for (int i = 0; i < mSections.Count; i++)
            {
                if ((mSections[i] as MIFSection).MinY < l_min_Y) l_min_Y = (mSections[i] as MIFSection).MinY;
            }
            return l_min_Y;
        }

        private double GetMaxX()
        {
            double l_max_X = double.MinValue;
            for (int i = 0; i < mSections.Count; i++)
            {
                if ((mSections[i] as MIFSection).MaxX > l_max_X) l_max_X = (mSections[i] as MIFSection).MaxX;
            }
            return l_max_X;
        }

        private double GetMaxY()
        {
            double l_max_Y = double.MinValue;
            for (int i = 0; i < mSections.Count; i++)
            {
                if ((mSections[i] as MIFSection).MaxY > l_max_Y) l_max_Y = (mSections[i] as MIFSection).MaxY;
            }
            return l_max_Y;
        }

        
        public override bool Draw(Bitmap aBitmap)
        {
            for (int i = 0; i < mSections.Count; i++)
            {
                (mSections[i] as MIFSection).MIFPen = mMIFPen;
                (mSections[i] as MIFSection).MIFBrush = mMIFBrush;
                
                if (!(mSections[i] as BaseObject).Draw(aBitmap)) return false;
            }
            return true;
        }

        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();
            sb.Append(string.Format("{0} {1}\r\n",MIFWords.Region,mSections.Count));
            for (int i = 0; i < mSections.Count; i++)
            {
                sb.Append((mSections[i] as MIFSection));
            }
            sb.Append(mMIFPen);
            sb.Append(mMIFBrush);
            if (mCenter.IsRead)
             sb.Append(string.Format("{0,10} {1}\r\n",MIFWords.Center,mCenter));
            return sb.ToString();
        }

       
        public MIFRegion(ArrayList aPolygons,MIFPen aPen,MIFBrush aBrush,MIFNode aCenter)
        {
            if (aCenter == null)
                mCenter = CalcCentroid();
            
                
            if (mMIFPen == null) mMIFPen = new MIFPen();
            if (mMIFBrush == null) mMIFBrush = new MIFBrush();
            
            mSections = aPolygons;
            mMIFPen      = aPen;
            mMIFBrush    = aBrush;
            mCenter   = aCenter;
            
            
            
            mMinX =  GetMinX();
            mMinY = GetMinY();
            mMaxX = GetMaxX();
            mMaxY = GetMaxY();

        }

        public static MIFRegion Parse(string aString)
        {
            MIFPen pen = null;
            MIFBrush brush = null;
            MIFNode center = null;
            try
            {
                if (aString == null) throw new ApplicationException("Invalid Region string");
                int current_token = 0;
                string[] tokens = Strings.Split(aString,new char[] { ' ', '\t', ',', '(', ')', '\n', '\r' });
                if (string.Compare(tokens[current_token], MIFWords.Region, true) != 0) throw new ApplicationException("Invalid Region string");
                current_token++;
                int sections_count = int.Parse(tokens[current_token]);
                ArrayList polygons = new ArrayList(sections_count);
                current_token++;
                for (int i = 0; i < sections_count; i++)
                {
                    int num_points = int.Parse(tokens[current_token]);
                    ArrayList nodes = new ArrayList(num_points);
                    current_token++;
                    for (int j = 0; j < num_points; j++)
                    {
                        MIFNode node = new MIFNode(double.Parse(tokens[current_token]), double.Parse(tokens[current_token + 1]), true);
                        nodes.Add(node);
                        current_token = current_token + 2;
                    }
                    MIFSection section = new MIFSection(nodes,SectionType.Polygon,new MIFPen(),new MIFBrush());
                    polygons.Add(section);
                }
                if (current_token < tokens.Length)
                    if (string.Compare(tokens[current_token], MIFWords.Pen, true) == 0)
                    {
                        current_token++;
                        pen = new MIFPen(int.Parse(tokens[current_token]), int.Parse(tokens[current_token + 1]), int.Parse(tokens[current_token + 2]));
                        current_token = current_token + 3;
                    }
                if (current_token < tokens.Length)
                    if (string.Compare(tokens[current_token], MIFWords.Brush) == 0)
                    {
                        current_token++;
                        brush = new MIFBrush(int.Parse(tokens[current_token]), int.Parse(tokens[current_token + 1]), int.Parse(tokens[current_token + 2]));
                        current_token = current_token + 3;
                    }
                if (current_token < tokens.Length)
                    if (string.Compare(tokens[current_token], MIFWords.Center) == 0)
                    {
                        current_token++;
                        center = new MIFNode(double.Parse(tokens[current_token]), double.Parse(tokens[current_token + 1]), true);
                        current_token = current_token + 2;
                    }

                for (int i = 0; i < polygons.Count; i++)
                {
                    (polygons[i] as MIFSection).MIFPen = pen;
                }

                    return new MIFRegion(polygons, pen, brush, center);
            }
            catch (Exception e)
            {
                MessageBox.Show(e.Message);
                return null;
            }


            

        }

      


        private MIFNode CalcCentroid()
        {
            double sum_x=0;
            double sum_y=0;
            int cnt_nodes = 0;
            for (int i = 0; i < mSections.Count;i++ )
            {
                MIFSection section = mSections[i] as MIFSection;
                for (int j = 0; j < section.Nodes.Count;j++ )
                {
                    MIFNode node = section.Nodes[j] as MIFNode;
                    cnt_nodes++;
                    sum_x += node.X;
                    sum_y += node.Y;
                }
            }
            return new MIFNode(sum_x / cnt_nodes, sum_y / cnt_nodes,false);
        }
    }
   
    #region OTHER_PRIMITIVES
    /*
    public class MIFArc:BaseObject
    {
        private MIFNode mNode1;
        private MIFNode mNode2;
        private MIFNode mDegrees;
        private MIFPen  mPen;


        public override bool IsSelected
        {
            get { return mIsSelected; }
            set { mIsSelected = value; }
        }


        public override bool DoSelect(double x, double y, double select_distance)
        {
            throw new Exception("The method or operation is not implemented.");
        }

        public override bool Draw(Bitmap aBitmap)
        {
            throw new ApplicationException("MIFArc.Draw() not implemented!");
        }

        public MIFArc(MIFNode aNode1,MIFNode aNode2,MIFNode aDegrees,MIFPen aPen)
        {
            mNode1   = aNode1; 
            mNode2   = aNode2;
            mDegrees = aDegrees;
            mPen     = aPen;
            if (mPen == null) mPen = new MIFPen();
        }

        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();
            sb.Append(string.Format("{0} {1} {2}\r\n",MIFWords.Arc,mNode1,mNode2));
            sb.Append(string.Format("{0}\r\n",mDegrees));
            if (mPen.IsRead)
            {
                sb.Append(mPen);
            }

            return sb.ToString();
        }

        public static MIFArc Parse(string aString)
        {
            try
            {
                if (aString == null) throw new ApplicationException("Invalid Arc string" + aString);

                string[] tokens = Strings.SplitAndRemoveEmpty(aString,new char[] { ' ', '\t', ',', '(', ')', '\n', '\r' });
                if (string.Compare(tokens[0], MIFWords.Arc, true) != 0) throw new ApplicationException("Invalid Arc string");

                if ((tokens.Length != 7) && (tokens.Length != 11)) throw new ApplicationException("Invalid Arc string");


                MIFArc result = new MIFArc(new MIFNode(Double.Parse(tokens[1]), Double.Parse(tokens[2]), true), new MIFNode(Double.Parse(tokens[3]), Double.Parse(tokens[4]), true), new MIFNode(Double.Parse(tokens[5]), Double.Parse(tokens[6]), true), new MIFPen());
                if ((tokens.Length == 11) && (string.Compare(tokens[7], MIFWords.Pen) == 0))
                    result.mPen = new MIFPen(int.Parse(tokens[8]), int.Parse(tokens[9]), int.Parse(tokens[10]), true);
                return result;
            }
            catch
            {
                throw new ApplicationException("Invalid Arc string" + aString);
            }

        }
    
    }
    
    public class MIFText : BaseObject
    {
        private string     mText;
        private MIFNode    mNode1;
        private MIFNode    mNode2;
        
        
        public override bool IsSelected
        {
            get { return mIsSelected; }
            set { mIsSelected = value; }
        }

        public override bool DoSelect(double x, double y, double select_distance)
        {
            throw new Exception("The method or operation is not implemented.");
        }

        public override bool Draw(Bitmap aBitmap)
        {
            throw new ApplicationException("MIFText.Draw() not implemented!");
        }

        public MIFText(string aText,MIFNode aNode1,MIFNode aNode2)
        {
            mText  = aText;
            mNode1 = aNode1;
            mNode2 = aNode2;
        }


        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();
            sb.Append(string.Format("{0} {1}\r\n {2} {3}", MIFWords.Text, mText, mNode1, mNode2));
            return sb.ToString();
        }

        public static MIFText Parse(string aString)
        {
            try
            {
                if (aString == null) throw new ApplicationException("Invalid Text string");

                string[] tokens =Strings.SplitAndRemoveEmpty(aString,new char[] { ' ', '\t', ',', '(', ')', '\n', '\r' });
                if (string.Compare(tokens[0], MIFWords.Text, true) != 0) throw new ApplicationException("Invalid Text string");

                return new MIFText(tokens[1], new MIFNode(double.Parse(tokens[2]), double.Parse(tokens[3]), true), new MIFNode(double.Parse(tokens[4]), double.Parse(tokens[5]), true));
            }
            catch
            {
                throw new ApplicationException("Invalid Text string"+aString);
            }
        }

    }
    
    public class MIFRectangle : BaseObject
    {
        private MIFPen   mPen;
        private MIFBrush mBrush;
        private MIFNode  mNode1;
        private MIFNode  mNode2;

        public override bool IsSelected
        {
            get { return mIsSelected; }
            set { throw new Exception("The method or operation is not implemented."); }
        }
        public override bool DoSelect(double x, double y, double select_distance)
        {
            throw new Exception("The method or operation is not implemented.");
        }

        public override bool Draw(Bitmap aBitmap)
        {
            Graphics gr = Graphics.FromImage(aBitmap);
            try
            {
                int x, y, x1, y1;
                Pen pen = new Pen(Color.Black);
                PointF p1 = new PointF(0, 0);
                PointF p2 = new PointF(0, 0);

                MapTrans.GetPixel(mNode1.X, mNode1.Y, out x, out y);
                MapTrans.GetPixel(mNode1.X, mNode1.Y, out x1, out y1);
                p1.X = x; p1.Y = y;
                p2.X = x1; p2.Y = y1;
                gr.DrawRectangle(pen, p1.X, p1.Y, p2.X, p2.Y);
                return true;
            }
            catch
            {
                return false;
            }
            finally
            {
                gr.Dispose();
            }
            
        } 
      
        public MIFRectangle(MIFNode aNode1,MIFNode aNode2,MIFPen aPen,MIFBrush aBrush)
        {
            mNode1 =  aNode1;
            mNode2 =  aNode2;
            mPen   =  aPen;
            mBrush =  aBrush;
            if (mPen == null) mPen = new MIFPen();
            if (mBrush == null) mBrush = new MIFBrush();
        }

        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();
            sb.Append(string.Format("{0} {1} {2}\r\n",MIFWords.Line,mNode1,mNode2));
            if (mPen.IsRead)
             sb.Append(mPen);
            if (mBrush.IsRead)
             sb.Append(mBrush);
            return sb.ToString();
        }

        public static MIFRectangle Parse(string aString)
        {

            try
            {
                if (aString == null) throw new ApplicationException("Invalid Rect string");

                string[] tokens = Strings.SplitAndRemoveEmpty(aString, new char[] { ' ', '\t', ',', '(', ')', '\n', '\r' });
                if (string.Compare(tokens[0], MIFWords.Rect, true) != 0) throw new ApplicationException("Invalid Rect string");
                MIFPen pen = null;
                MIFBrush brush = null;
                int current_token = 5;
                while (current_token < tokens.Length)
                {
                    if (string.Compare(tokens[current_token], MIFWords.Pen) == 0)
                    {
                        pen = new MIFPen(int.Parse(tokens[++current_token]), int.Parse(tokens[++current_token]), int.Parse(tokens[++current_token]), true);

                    }
                    if (string.Compare(tokens[current_token], MIFWords.Brush) == 0)
                    {
                        brush = new MIFBrush(int.Parse(tokens[++current_token]), int.Parse(tokens[++current_token]), int.Parse(tokens[++current_token]), true);

                    }
                    current_token++;
                }
                MIFRectangle result = new MIFRectangle(new MIFNode(Double.Parse(tokens[1]), Double.Parse(tokens[2]), true), new MIFNode(Double.Parse(tokens[3]), Double.Parse(tokens[4]), true), pen, brush);
                return result;
            }
            catch
            {
                throw new ApplicationException("Invalid Rectangle string"+aString);
            }
        }
    }
    
    public class MIFRndRectangle : BaseObject
    {
        private double mDegree;
        private MIFPen mPen;
        private MIFBrush mBrush;
        private MIFNode mNode1;
        private MIFNode mNode2;

        public override bool IsSelected
        {
            get { return mIsSelected; }
            set { throw new Exception("The method or operation is not implemented."); }
        }

        public override bool DoSelect(double x, double y, double select_distance)
        {
            throw new Exception("The method or operation is not implemented.");
        }

        public override bool Draw(Bitmap aBitmap)
        {
            throw new ApplicationException("RndRect.Draw() not implemented!");
        }
        public MIFRndRectangle(MIFNode aNode1,MIFNode aNode2,double aDegree,MIFPen aPen,MIFBrush aBrush)
        {

            mNode1 = aNode1;
            mNode2 = aNode2;
            mDegree = aDegree;
            mPen = aPen;
            mBrush = aBrush;
            if (mPen == null) mPen = new MIFPen();
            if (mBrush== null) mBrush = new MIFBrush();
        }
        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();
            sb.Append(string.Format("{0} {1} {2}\r\n",MIFWords.RoundRect,mNode1,mNode2));
            sb.Append(string.Format("{0}\r\n",mDegree)); 
            if (mPen.IsRead)
             sb.Append(mPen);
            if (mBrush.IsRead)
             sb.Append(mBrush);
            return sb.ToString();
        }
        public static MIFRndRectangle Parse(string aString)
        {
            try
            {

                if (aString == null) throw new ApplicationException("Invalid RndRect string");
                string[] tokens = Strings.SplitAndRemoveEmpty(aString, new char[] { ' ', '\t', ',', '(', ')', '\n', '\r' });
                if (string.Compare(tokens[0], MIFWords.RoundRect, true) != 0) throw new ApplicationException("Invalid RndRect string");
                MIFPen pen = null;
                MIFBrush brush = null;
                int current_token = 6;
                while (current_token < tokens.Length)
                {
                    if (string.Compare(tokens[current_token], MIFWords.Pen) == 0)
                    {
                        pen = new MIFPen(int.Parse(tokens[++current_token]), int.Parse(tokens[++current_token]), int.Parse(tokens[++current_token]), true);

                    }
                    if (string.Compare(tokens[current_token], MIFWords.Brush) == 0)
                    {
                        brush = new MIFBrush(int.Parse(tokens[++current_token]), int.Parse(tokens[++current_token]), int.Parse(tokens[++current_token]), true);

                    }
                    current_token++;
                }
                MIFRndRectangle result = new MIFRndRectangle(new MIFNode(Double.Parse(tokens[1]), Double.Parse(tokens[2]), true), new MIFNode(Double.Parse(tokens[3]), Double.Parse(tokens[4]), true), Double.Parse(tokens[5]), pen, brush);
                return result;
            }
            catch
            {
                throw new ApplicationException("Invalid RndRect string");
            }
        }

    }
    
    public class MIFEllipse : BaseObject
    {
        private MIFPen mPen;
        private MIFBrush mBrush;

        private MIFNode mNode1;
        private MIFNode mNode2;
        
        public override bool IsSelected
        {
            get { return mIsSelected; }
            set { throw new Exception("The method or operation is not implemented."); }
        }

        public override bool DoSelect(double x, double y, double select_distance)
        {
            throw new Exception("The method or operation is not implemented.");
        }

        public override bool Draw(Bitmap aBitmap)
        {
            Graphics gr = Graphics.FromImage(aBitmap);
            int x1,y1,x2,y2;
            MapTrans.GetPixel(mNode1.X, mNode1.Y, out x1, out y1);
            MapTrans.GetPixel(mNode2.X, mNode2.Y, out x2, out y2);
            Pen pen = new Pen(Color.Black);
            Rectangle rect = new Rectangle((int)x1,(int)y1,(int)(x2-x1),(int)(y2-y1));
            gr.DrawEllipse(pen, rect);
            gr.Dispose();
            return true;
        }

        public MIFEllipse(MIFNode aNode1,MIFNode aNode2,MIFPen aPen,MIFBrush aBrush)
        {
            mNode1 = aNode1;
            mNode2 = aNode2;
            mPen   = aPen;
            mBrush = aBrush;
            if (mPen == null) mPen = new MIFPen();
            if (mBrush== null) mBrush = new MIFBrush();
        }

        public override string ToString()
        {
            
            StringBuilder sb = new StringBuilder(MIFWords.Ellipse); sb.Append(" ");
            sb.Append(mNode1); sb.Append(" ");
            sb.Append(mNode2); sb.Append(" ");
            sb.Append("\r\n");
            if (mPen.IsRead)
                sb.Append(mPen);
            if (mBrush.IsRead) 
                sb.Append(mBrush);
            return sb.ToString();
        }

        public static MIFEllipse Parse(string aString)
        {
            try
            {

                if (aString == null) throw new ApplicationException("Invalid Ellipse string");

                string[] tokens = Strings.SplitAndRemoveEmpty(aString, new char[] { ' ', '\t', ',', '(', ')', '\n', '\r' });
                if (string.Compare(tokens[0], MIFWords.Ellipse, true) != 0) throw new ApplicationException("Invalid Ellipse string");
                MIFPen pen = null;
                MIFBrush brush = null;
                int current_token = 5;
                while (current_token < tokens.Length)
                {
                    if (string.Compare(tokens[current_token], MIFWords.Pen) == 0)
                    {
                        pen = new MIFPen(int.Parse(tokens[++current_token]), int.Parse(tokens[++current_token]), int.Parse(tokens[++current_token]), true);
                    }
                    if (string.Compare(tokens[current_token], MIFWords.Brush) == 0)
                    {
                        brush = new MIFBrush(int.Parse(tokens[++current_token]), int.Parse(tokens[++current_token]), int.Parse(tokens[++current_token]), true);
                    }
                    current_token++;
                }
                MIFEllipse result = new MIFEllipse(new MIFNode(Double.Parse(tokens[1]), Double.Parse(tokens[2]), true), new MIFNode(Double.Parse(tokens[3]), Double.Parse(tokens[4]), true), pen, brush);
                return result;
            }
            catch
            {
                throw new ApplicationException("Invalid Ellipse string");
            }
        }

    }
        
    public class MIFMultipoint : BaseObject
        {
            private ArrayList mNodes;
            private MIFSymbol mSymbol;

        public override bool IsSelected
        {
            get { return mIsSelected; }
            set { throw new Exception("The method or operation is not implemented."); }
        }    

            public override bool DoSelect(double x, double y, double select_distance)
             {
               throw new Exception("The method or operation is not implemented.");
             }    

            public override bool Draw(Bitmap aBitmap)
            {
                throw new ApplicationException("MultiPoint.Draw() not implemented!");
            }
            
            public MIFMultipoint(ArrayList aNodes,MIFSymbol aSymbol)
            {
                mNodes  = aNodes;
                mSymbol = aSymbol;
                if (mSymbol == null) mSymbol = new MIFSymbol();
            }

            public override string ToString()
            {
                StringBuilder sb = new StringBuilder(MIFWords.MultiPoint); sb.Append(" ");sb.Append(mNodes.Count.ToString());
                sb.Append("\r\n");                
                for (int i = 0;i<mNodes.Count;i++)
                {
                    sb.Append((mNodes[i] as MIFNode));
                    sb.Append("\r\n");
                }
                if (mSymbol.IsRead)
                    sb.Append(mSymbol);
                return sb.ToString();
            }

            public static MIFMultipoint Parse(string aString)
            {
                try
                {
                    ArrayList nodes = new ArrayList();
                    MIFSymbol symbol = null;

                    if (aString == null) throw new ApplicationException("Invalid Multipoint string");

                    string[] tokens = Strings.SplitAndRemoveEmpty(aString, new char[] { ' ', '\t', ',', '(', ')', '\n', '\r' });
                    if (string.Compare(tokens[0], MIFWords.MultiPoint, true) != 0) throw new ApplicationException("Invalid Multipoint string");

                    int nodes_count = int.Parse(tokens[1]);
                    int current_token = 1;
                    for (int i = 0; i < nodes_count; i++)
                    {
                        MIFNode node = new MIFNode(double.Parse(tokens[++current_token]), double.Parse(tokens[++current_token]), true);
                        nodes.Add(node);
                    }
                    current_token++;
                    while (current_token < tokens.Length)
                    {
                        if (string.Compare(tokens[current_token], MIFWords.Symbol, true) == 0)
                        {
                            symbol = new MIFSymbol(int.Parse(tokens[++current_token]), int.Parse(tokens[++current_token]), int.Parse(tokens[++current_token]), true);

                        }
                        current_token++;
                    }
                    return new MIFMultipoint(nodes, symbol);
                }
                catch
                {
                    throw new ApplicationException("Invalid Multipoint string! "+aString);
                }
            }
        }
       
    public class MIFCollection : BaseObject
            {
                private MIFRegion mRegion;
                private MIFPolyLine mPline;
                private MIFMultipoint mMultipoint;
                private static bool IsObjectType(string aString)
                {
                    return
                        (string.Compare(MIFWords.Region, aString, true) == 0) ||
                        (string.Compare(MIFWords.Pline, aString, true) == 0)  ||
                        (string.Compare(MIFWords.MultiPoint, aString, true) == 0);    

                }

        public override bool IsSelected
        {
            get { return mIsSelected; }
            set { throw new Exception("The method or operation is not implemented."); }
        }

                public override bool DoSelect(double x, double y, double select_distance)
                    {
                        throw new Exception("The method or operation is not implemented.");
                    }
                public override bool Draw(Bitmap aBitmap)
                {
                    throw new ApplicationException("Collection.Draw() not implemented!");
                }
                
                public MIFCollection(MIFRegion aRegion,MIFPolyLine aPolyLine,MIFMultipoint aMultipoint)
                {
                    mRegion     = aRegion;
                    mPline      = aPolyLine;
                    mMultipoint = aMultipoint;
                }

                public override string ToString()
                {
                    StringBuilder sb = new StringBuilder(MIFWords.Collection);sb.Append(" ");
                    int num_parts = 3;
                    if (mRegion == null) num_parts--;
                    if (mMultipoint == null) num_parts--;
                    if (mPline == null) num_parts--;
                    sb.Append(num_parts.ToString()); sb.Append("\r\n");
                    if (mRegion != null) sb.Append(mRegion);
                    if (mPline != null) sb.Append(mPline);
                    if (mMultipoint != null) sb.Append(mMultipoint);
                    return sb.ToString();
                }

                public static MIFCollection Parse(string aString)
                {
                    try
                    {
                        if ((aString == null) || (aString.Trim() == "")) throw new ApplicationException("Invalid Collection string");
                        string[] tokens = Strings.SplitAndRemoveEmpty(aString, new char[] { ' ', '\t', ',', '(', ')', '\n', '\r' });
                        if (string.Compare(tokens[0], MIFWords.Collection, true) != 0) throw new ApplicationException("Invalid Collection string");
                        MIFRegion region = null;
                        MIFPolyLine polyline = null;
                        MIFMultipoint multipoint = null;
                        int current_token = 1;
                        if (!IsObjectType(tokens[current_token])) current_token++;

                        while (current_token < tokens.Length)
                        {
                            StringBuilder sb = new StringBuilder(tokens[current_token]);
                            string str_obj = tokens[current_token];
                            current_token++;
                            sb.Append(" ");
                            while ((current_token < tokens.Length) && (!IsObjectType(tokens[current_token])))
                            {
                                sb.Append(tokens[current_token]);
                                sb.Append(" ");
                                current_token++;
                            }
                            if (string.Compare(MIFWords.Region, str_obj, true) == 0)
                                region = MIFRegion.Parse(sb.ToString());
                            if (string.Compare(MIFWords.Pline, str_obj, true) == 0)
                                polyline = MIFPolyLine.Parse(sb.ToString());
                            if (string.Compare(MIFWords.MultiPoint, str_obj, true) == 0)
                                multipoint = MIFMultipoint.Parse(sb.ToString());
                        }
                        return new MIFCollection(region, polyline, multipoint);
                    }
                    catch
                    {
                        throw new ApplicationException("Invalid Collection string");
                    }

                }
            }
    */
    #endregion

    public class MIFSection:BaseObject
    {
        private ArrayList mNodes;
        private MIFPen m_MIFPen;
        private MIFBrush m_MIFBrush;
        private SectionType m_SectionType;


        public void Clear()
        {
            mNodes.Clear();
        }
        public MIFPen   MIFPen 
            {
            set { m_MIFPen = value; }
            get { return m_MIFPen; }
            }
        public MIFBrush MIFBrush {
            get { return m_MIFBrush; }
            set { m_MIFBrush = value; } 
        }

        public void AddNode(MIFNode aNode)
        {
            mNodes.Add(aNode);
            mMinX = Math.Min(mMinX, aNode.X);
            mMinY = Math.Min(mMinY, aNode.Y);
            mMaxX = Math.Max(mMaxX, aNode.X);
            mMaxY = Math.Max(mMaxY, aNode.Y);
        }

        public void RemoveAt(int aIdx)
        {
            mNodes.RemoveAt(aIdx);
            if (mNodes.Count <= 1) mNodes.Clear();
        }
        

        
        public override int GetHashCode()
        {
            int result = base.GetHashCode();
            for (int i = 0; i < mNodes.Count; i++)
            {
                result = result ^ mNodes[i].GetHashCode();
            }
            return result;
        }

        public override bool Equals(object obj)
        {
            if (obj == null || GetType() != obj.GetType())
                return false;
            if (this.mNodes.Count!=(obj as MIFSection).mNodes.Count ) return false;

            for (int i = 0; i < mNodes.Count; i++)
            {
                if (!(mNodes[i].Equals((obj as MIFSection).mNodes[i]))) return false;
            }
            return true;
        }

        public override bool IsSelected
        {
            get { return mIsSelected; }
            set 
            {
                mIsSelected = value;   
            }
        }



        public override bool DoSelect(double x, double y, double select_distance)
        {
            if (m_SectionType == SectionType.Polygon)
            {
                bool result = false;
                int i, j = 0;
                for (i = 0; i < mNodes.Count; i++)
                {
                    j++; if (j == mNodes.Count) j = 0;

                    MIFNode nI = (MIFNode)mNodes[i];
                    MIFNode nJ = (MIFNode)mNodes[j];

                    if ((nI.Y < y) && (nJ.Y >= y) || (nJ.Y < y) && (nI.Y >= y))
                    {
                        if (nI.X + (y - nI.Y) / (nJ.Y - nI.Y) * (nJ.X - nI.X) < x)
                        {
                            result = !result;
                        }
                    }
                }

                return result;
            }
            else //PolyLine
            {
                for (int i = 0; i < mNodes.Count-1; i++)
                {
                    if (Geometry.DistDegLineToDegPointInPixels((mNodes[i] as MIFNode).X, (mNodes[i] as MIFNode).Y,
                        (mNodes[i + 1] as MIFNode).X, (mNodes[i+1] as MIFNode).Y, x, y, MapTrans.mPixelsPerDegree) <= select_distance) return true;             
                }
                return false;
            }
        }

        
        public MIFSection(ArrayList aNodes,SectionType aSectionType,MIFPen aMIFPen,MIFBrush aMIFBrush)
        {
            mNodes = aNodes;
            mMinX = GetMinX();
            mMinY = GetMinY();
            mMaxX = GetMaxX();
            mMaxY = GetMaxY();
            m_SectionType = aSectionType;
            m_MIFPen = aMIFPen;
            m_MIFBrush = aMIFBrush;
        }

        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();
            sb.Append("  "+mNodes.Count.ToString() + "\n\r");
            for (int i = 0; i < mNodes.Count; i++)
            {
                sb.Append(mNodes[i].ToString());
                sb.Append("\n\r");
            }
            return sb.ToString();
        }

        public ArrayList Nodes { get { return mNodes; } }

        public override bool Draw(Bitmap aBitmap)
        {
            Graphics gr = null;
            Pen pen = null;
            Brush brush = null;
            int x, y;
            try
            {
                gr = Graphics.FromImage(aBitmap);
                Point[] points = new Point[mNodes.Count];

                for (int i = 0; i < mNodes.Count; i++)
                {
                    MapTrans.GetPixel((mNodes[i] as MIFNode).X, (mNodes[i] as MIFNode).Y, out x, out y);
                    Point p = new Point(x, y);
                    points[i] = p;

                }

                if (mIsSelected)
                {
                    
                     pen = new Pen(Defaults.SelectColor,MIFPen.Width-1);
                     brush = m_MIFBrush.NewBrush();
                    if (m_SectionType == SectionType.Polygon)
                        gr.FillPolygon(brush, points);
                    if (points.Length > 1)
                        gr.DrawLines(pen, points);
                }
                else
                {
                    pen = m_MIFPen.NewPen();
                    brush = m_MIFBrush.NewBrush();
                    if (m_SectionType == SectionType.Polygon)
                        gr.FillPolygon(brush, points);
                    if (points.Length > 1)
                        gr.DrawLines(pen, points);
                }
            }
            catch { return false; }
            finally
            {
                if (brush!=null)
                    brush.Dispose();
                if (pen!=null)
                    pen.Dispose();
                gr.Dispose();
            }
            return true;
        }

        public static MIFSection Parse(string aString,SectionType aSectionType)
        {
            if (aString == null) throw new ApplicationException("Invalid Section string");
            string[] tokens = aString.Split(new char[] { ' ', '\t', ',', '(', ')', '\n', '\r' }, StringSplitOptions.RemoveEmptyEntries);
            int num_nodes = int.Parse(tokens[0]);
            int current_token = 1;
            ArrayList nodes = new ArrayList(num_nodes);
            for (int i = 0; i < num_nodes; i++)
            {
                MIFNode node = new MIFNode(Double.Parse(tokens[current_token++]), Double.Parse(tokens[current_token++]),true);
                nodes.Add(node);
            }

            return new MIFSection(nodes,aSectionType,new MIFPen(),new MIFBrush());

        }

        private double GetMinX()
        {
            double l_min_X = double.MaxValue;
            for (int i = 0; i < mNodes.Count; i++)
            {
                if ((mNodes[i] as MIFNode).X < l_min_X) l_min_X = (mNodes[i] as MIFNode).X;
            }
            return l_min_X;
        }
        private double GetMinY()
        {
            double l_min_Y = double.MaxValue;
            for (int i = 0; i < mNodes.Count; i++)
            {
                if ((mNodes[i] as MIFNode).Y < l_min_Y) l_min_Y = (mNodes[i] as MIFNode).Y;
            }
            return l_min_Y;
        }

        private double GetMaxX()
        {
            double l_max_X = double.MinValue;
            for (int i = 0; i < mNodes.Count; i++)
            {
                if ((mNodes[i] as MIFNode).X > l_max_X) l_max_X = (mNodes[i] as MIFNode).X;
            }
            return l_max_X;
        }
        private double GetMaxY()
        {
            double l_max_Y = double.MinValue;
            for (int i = 0; i < mNodes.Count; i++)
            {
                if ((mNodes[i] as MIFNode).Y > l_max_Y) l_max_Y = (mNodes[i] as MIFNode).Y;
            }
            return l_max_Y;
        }


    }
    
    public class MIFNode
    {
        private double mX;
        private double mY;
        private bool mIsRead;

        public override int GetHashCode()
        {
            return mX.GetHashCode() ^ mY.GetHashCode();
        }
        
        public override bool Equals(object obj)
        {
            if (obj == null || GetType() != obj.GetType())
                return false;

            return
                    (
                    (this.mX == (obj as MIFNode).mX) &&
                    (this.mY == (obj as MIFNode).mY)
                    );
        }
             
        
        public bool IsRead
        {
            get { return mIsRead; }
        }
    
        public MIFNode(double aX,double aY,bool IsRead)
        {
            mX = aX;
            mY = aY;
            mIsRead = IsRead;
        }
        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();
            sb.Append(string.Format("{0} {1}", mX, mY));
            return sb.ToString();
        }
        public double X { get { return mX; } }
        public double Y { get { return mY; } }
   
    }

    public sealed class MIFEmpty:BaseObject
    {
        public override bool IsSelected
        {
            get { return mIsSelected; }
            set { mIsSelected = false; }
        }

        public override bool Draw(Bitmap aBitmap)
        {
            return true;
        }
        public override bool DoSelect(double x, double y, double select_distance)
        {
            return false;
        }
    }

    public class MIFSymbol
    {
        public const int MinSize = 1;
        public const int MaxSize = 48;

        private int mShape;
        private int mColor;
        private int mSize;
       
        public static Bitmap mSymbols =new Bitmap(@"SYMBOLS.bmp");


         
        
        
        public static void DrawSymbol(Graphics gr,int aShape)
        {
             gr.FillRectangle(Brushes.White, gr.ClipBounds);
            int x = (desc_symbol_s.GetByIndex(aShape-31) as DESC_BITMAP).X1+3;
            int y = (desc_symbol_s.GetByIndex(aShape-31) as DESC_BITMAP).Y1+4;
            int width = (desc_symbol_s.GetByIndex(aShape-31) as DESC_BITMAP).WIDTH-2;
            int height = (desc_symbol_s.GetByIndex(aShape-31) as DESC_BITMAP).HEIGHT-2;
            Rectangle rect = new Rectangle(x,y,width,height);
            gr.DrawImage(mSymbols,0,0,rect,GraphicsUnit.Pixel);
        }
        public static DESC_SYMBOL_S desc_symbol_s = new DESC_SYMBOL_S();

        public int Shape 
        { 
            get { return mShape; } 
            set {mShape = value; }
        }
        public int Color
        {
            get { return mColor; }
            set { mColor = value;}
        }
        public int Size
        {
            get { return mSize; }
            set { mSize = value;}
        }


        public  static void PutToComboShape(ComboBox.ObjectCollection aItems)
        {
            desc_symbol_s.PutToCombo(aItems);
        }

        public static void PutToComboSize(ComboBox.ObjectCollection aItems)
        {
            aItems.Clear();
            for (int i = MinSize; i < MaxSize; i++)
            {
                MyUtils.Strings.ItemsAdd(aItems,i.ToString(),i.ToString());
            }
        }

        public MIFSymbol(int aShape, int aColor, int aSize)
        {
            mShape  = aShape;
            mColor  = aColor;
            mSize   = aSize;
        }

        public override string ToString()
        {
            return  string.Format("{0} ({1},{2},{3})\r\n",MIFWords.Symbol,mShape,mColor,mSize);
        }

        public MIFSymbol()
            : this(Defaults.Symbol_Shape, Defaults.Symbol_Color, Defaults.Symbol_Size)
        {
        }


        public bool Draw(int x,int y,Graphics gr,bool IsSelected)
        {
            if (mShape == 31) return true;
                DESC_BITMAP ds = desc_symbol_s.GetByIndex(desc_symbol_s.IndexOfKey(mShape)) as DESC_BITMAP;
                Rectangle rect_source = new Rectangle(ds.X1,ds.Y1,ds.WIDTH,ds.HEIGHT);
                Rectangle rect_dest   = new Rectangle(x,y,mSize,mSize);
                gr.DrawImage(mSymbols,rect_dest,rect_source,GraphicsUnit.Pixel);
                if (IsSelected)
                {
                    rect_dest.Height--;
                    rect_dest.Width--;
                    Pen pen = new Pen(Defaults.SelectColor);
                    gr.DrawRectangle(pen, rect_dest);
                    pen.Dispose();
                    //Pen pen = Pens.Black;
                    //gr.DrawRectangle(pen, rect_dest);
                    //Rectangle rect_vert = new Rectangle();
                    //rect_vert.X = rect_dest.X - 3;
                    //rect_vert.Y = rect_dest.Y - 3;
                    //rect_vert.Width = rect_vert.Height = 6;
                    //gr.DrawRectangle(pen,rect_vert);
                }
            return true;
        }

    


    }

    public class MIFPen
    {
        private int mWidth;
        private int mPattern;
        private int mColor;

        public MIFPen()
            : this(Defaults.Pen_width, Defaults.Pen_pattern, Defaults.Pen_color)
        {

        }

        public MIFPen(MIFPen aMIFPen)
            :this(aMIFPen.mWidth,aMIFPen.mPattern,aMIFPen.mColor)
        {
        }
        public MIFPen(int aWidth,int aPattern, int aColor)
        {
            mWidth      = aWidth;
            mPattern    = aPattern;
            mColor      = aColor;
        }
        public Pen NewPen()
        {
            return new Pen(MyConvert.Rgb24ToColor(mColor), mWidth);
        }
    
        public int Width 
        {
            get { return mWidth; }
            set { mWidth = value; }
        }
        public int MIFColor
        {
            get { return mColor; }
            set { mColor = value; }
        }
        public Color Color
        {
            get { return MyConvert.Rgb24ToColor(mColor); }
        }

        public override string ToString()
        {
            return string.Format("{0,7} ({1},{2},{3})\r\n", MIFWords.Pen, mWidth, mPattern, mColor);
        }

    }

    public class MIFBrush
    {
        private int mPattern;
        private int mForeColor;
        private int mBackColor;


        public MIFBrush(int aPattern, int aForeColor, int aBackColor)
        {
            mPattern = aPattern;
            mForeColor = aForeColor;
            mBackColor = aBackColor;
        }
        public MIFBrush()
            : this(Defaults.Brush_pattern, Defaults.Brush_forecolor, Defaults.Brush_backcolor)
        {

        }
        public MIFBrush(MIFBrush aMIFBrush)
            : this(aMIFBrush.mPattern, aMIFBrush.mForeColor, aMIFBrush.mBackColor)
        {

        }

        
        public SolidBrush NewBrush()
        {
                       
            return new SolidBrush(MyConvert.Rgb24ToColor(mBackColor));
        }
        
        public int MIFBackColor 
        { 
            get { return mBackColor;  }
            set { mBackColor = value; }
        }

        public Color BackColor
        {
            get { return MyConvert.Rgb24ToColor(mBackColor); }
            
        }

        public override string ToString()
        {
            return string.Format("{0,9} ({1},{2},{3})\r\n", MIFWords.Brush, mPattern, mForeColor, mBackColor);
        }


        

 
    }


    public enum SectionType { Polygon, PolyLine }




}
