using System;
using MyUtils;
using System.Text;
using System.Collections;



namespace MyNMEA
{
    public enum NMEA { None,GGA, GLL, GSA, GSV, RMC, VTG }
    
    public class Misc
    {
        public const string VALID = "A";
        public static bool IsGGAMessage(string aString)
        {
            return (aString.StartsWith("$GPGGA"));
        }
        public static bool IsGLLMessage(string aString)
        {
            return (aString.StartsWith("$GPGLL"));
        }
        public static bool IsGSAMessage(string aString)
        {
            return (aString.StartsWith("$GPGSA"));
        }
        public static bool IsGSVMessage(string aString)
        {
            return (aString.StartsWith("$GPGSV"));
        }
        public static bool IsRMCMessage(string aString)
        {
            return (aString.StartsWith("$GPRMC"));
        }
        public static bool IsVTGMessage(string aString)
        {
            return (aString.StartsWith("$GPVTG"));
        }

        public static double NmeaFormatDegreesToDegrees(double aNmeaFormatDegrees,string aE_W_N_S)
        {
            int sign = 1;
            if (string.Compare(aE_W_N_S, Consts.c_South) == 0) sign = -1;
            if (string.Compare(aE_W_N_S, Consts.c_West) == 0)  sign = -1;
            return sign * (NmeaFormatDegreesToDegrees(aNmeaFormatDegrees));
        }

        public static double NmeaFormatDegreesToDegrees(double aNmeaFormatDegrees)
        {
            double ddd = Math.Truncate(aNmeaFormatDegrees / 100);
            double mm_mmm = Math.IEEERemainder(aNmeaFormatDegrees, 100);
            return (ddd + (mm_mmm / 60));
        }

    }
    
    
    public abstract class NMEA_Msg
    {
        public readonly int FieldsCount;
        protected string m_RawMessage;
        
        protected string m_Name;
        protected string m_NameStr
        {
            get { return m_Name; }
            set { m_Name = value; }
        }

        protected int m_CS;
        protected string m_CSStr
        {
            get { return string.Format("*{0}", m_CS.ToString("X2")); }
            set { m_CS = Convert.ToInt32(value, 16); }
        }

        public abstract string this[int index]
        {
            get;
            set;
        }
        protected static int CalcCheckSumStr(string aStrMsg)
        {
            int l_Checksum = 0;
            foreach (char Character in aStrMsg)
            {
                if (Character == '$') continue;
                else
                    if (Character == '*') break;
                    else
                    {
                        if (l_Checksum == 0)
                            l_Checksum = Convert.ToByte(Character);
                        else
                            l_Checksum = l_Checksum ^ Convert.ToByte(Character);
                    }
            }
            return l_Checksum;
        }
        protected static int GetCheckSum(string aStrMsg)
        {
            try
            {
                return Convert.ToInt32(aStrMsg.Substring(aStrMsg.IndexOf('*') + 1), 16);
            }
            catch
            {
                return 0;
            }
        }
        protected static bool IsCorrectCheckSum(string aStrMsg)
        {
            return (GetCheckSum(aStrMsg)==CalcCheckSumStr(aStrMsg));
        }
        public NMEA_Msg(int aFieldCount)
        {
            FieldsCount = aFieldCount;
        }
        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();
            for (int i = 0; i < FieldsCount; i++)
            {
                sb.Append(this[i]);
                if (i < FieldsCount - 2) sb.Append(',');
            }
            return sb.ToString();
        }
        public static bool IsCorrectNMEA(string aMsg,out NMEA aenENMA )
        {
            aenENMA = NMEA.None;
            if (GGA.IsCorrectGGA(aMsg) ) { aenENMA = NMEA.GGA; return true; }
            if (GLL.IsCorrectGLL(aMsg) ) { aenENMA = NMEA.GLL; return true; }
            if (GSA.IsCorrectGSA(aMsg) ) { aenENMA = NMEA.GSA; return true; }
            if (GSV.IsCorrectGSV(aMsg) ) { aenENMA = NMEA.GSV; return true; }
            if (RMC.IsCorrectRMC(aMsg) ) { aenENMA = NMEA.RMC; return true; }
            if (VTG.IsCorrectVTG(aMsg) ) { aenENMA = NMEA.VTG; return true; } 
            return false;
        }
        public abstract bool IsValid();
    }

    
    public sealed class GGA :   NMEA_Msg
    {

        private double m_UtcTime;
        private string m_UtcTimeStr 
        {
            get {
                return m_UtcTime.ToString();  
                }
            set {
                MyConvert.TryParseDouble(value, out m_UtcTime);
                }
        }

        private double  m_Latitude;
        private string  m_LatitudeStr
        {
            get {
                 return m_Latitude.ToString(); 
                }
            set {
                MyConvert.TryParseDouble(value, out m_Latitude);
                }
        }
        public  double  LatitudeDeg
        {
            
            get 
            {
                
                return Misc.NmeaFormatDegreesToDegrees(m_Latitude,m_NS_Indicator);
            }
        }

        private string m_NS_Indicator;
        private string m_NS_IndicatorStr
        {
            get { return m_NS_Indicator ;}
            set { m_NS_Indicator = value; }
        }
        
        private double  m_Longitude ;
        private string  m_LongitudeStr
        {
            get {
                    return m_Longitude.ToString();
                }
            set {
                MyConvert.TryParseDouble(value, out m_Longitude);
                }
        }
        public  double  LongitudeDeg
        {

            get
            {
                return Misc.NmeaFormatDegreesToDegrees(m_Longitude,m_NS_Indicator);
            }
        }

        private string m_EW_Indicator;
        private string m_EW_IndicatorStr
        {
            get { return m_EW_Indicator; }
            set { m_EW_Indicator = value; }
        }

        private UInt16 m_PosFixIndicator;
        private string m_PosFixIndicatorStr
        {
            get {
                    return m_PosFixIndicator.ToString(); 
                }
            set {
                    MyConvert.TryParseUInt16(value, out m_PosFixIndicator);
                }
        }

        private UInt16 m_SatellitesUsed;
        private string m_SatellitesUsedStr
        {
            get {
                    return string.Format("{0}",m_SatellitesUsed); 
                }
            set {
                MyConvert.TryParseUInt16(value, out m_SatellitesUsed);
                }
        }

        private double m_HDOP;
        private string m_HDOPStr
        {
            get { return m_HDOP.ToString(); }
            set { MyConvert.TryParseDouble(value, out m_HDOP); }
        }

        private double m_MSL_Altitude;
        private string m_MSL_AltitudeStr
        {
            get { return m_MSL_Altitude.ToString(); }
            set { MyConvert.TryParseDouble(value, out m_MSL_Altitude); }
        }

        private string m_MSL_Units;
        private string m_MSL_UnitsStr
        {
            get { return m_MSL_Units; }
            set { m_MSL_Units = value; }
        }

        private double m_AltRef;
        private string m_AltRefStr
        {
            get { return m_AltRef.ToString(); }
            set { MyConvert.TryParseDouble(value, out m_AltRef); }
        }

        private string m_AltRef_Units;
        private string m_AltRef_UnitsStr
        {
            get { return m_AltRef_Units; }
            set { m_AltRef_Units = value; }
        }

        private double m_DiffAge;
        private string m_DiffAgeStr
        {
            get { return m_DiffAge.ToString(); }
            set { MyConvert.TryParseDouble(value, out m_DiffAge); }
        }

        private double m_DiffStation;
        private string m_DiffStationStr
        {
            get { return m_DiffStation.ToString();}
            set { MyConvert.TryParseDouble(value, out m_DiffStation); }
        }

        public GGA(int aFieldsCount):base(aFieldsCount)
        {

        }
        public override string this[int index]
        {
            get 
            {
                switch (index)
                {
                    case 0: return m_NameStr;                        
                    case 1: return m_UtcTimeStr;
                    case 2: return m_LatitudeStr;
                    case 3: return m_NS_IndicatorStr;
                    case 4: return m_LongitudeStr;
                    case 5: return m_EW_IndicatorStr;
                    case 6: return m_PosFixIndicatorStr;
                    case 7: return m_SatellitesUsedStr;
                    case 8: return m_HDOPStr;
                    case 9: return m_MSL_AltitudeStr;
                    case 10: return m_MSL_UnitsStr;
                    case 11: return m_AltRefStr;
                    case 12: return m_AltRef_UnitsStr;
                    case 13: return m_DiffAgeStr;
                    case 14: return m_DiffStationStr;
                    case 15: return m_CSStr;
                    default: return "";
                }
            }
            set
            {
                switch (index)
                {
                    case 0: m_NameStr = value; break;
                    case 1: m_UtcTimeStr = value; break;
                    case 2: m_LatitudeStr = value; break;
                    case 3: m_NS_IndicatorStr = value; break;
                    case 4: m_LongitudeStr = value; break;
                    case 5: m_EW_IndicatorStr = value; break;
                    case 6: m_PosFixIndicatorStr = value; break;
                    case 7: m_SatellitesUsedStr = value; break;
                    case 8: m_HDOPStr = value; break;
                    case 9: m_MSL_AltitudeStr = value; break;
                    case 10: m_MSL_UnitsStr = value; break;
                    case 11: m_AltRefStr = value; break;
                    case 12: m_AltRef_UnitsStr = value; break;
                    case 13: m_DiffAgeStr = value; break;
                    case 14: m_DiffStationStr = value; break;
                    case 15: m_CSStr = value; break;
                }
            }
        }
        public static GGA Parse(string aMessage)
        {
            try
            {
                GGA msg = new GGA(16);
                msg.m_RawMessage = aMessage;
                string[] words = Strings.Split(msg.m_RawMessage, new char[] { ',', '*' }, false);
                for (int i = 0; i < words.Length; i++)
                {
                    if (words[i] == null) continue;
                    msg[i] = words[i];
                }
                return msg;
            }
            catch
            {
                return null;
            }
        }
        public static bool IsCorrectGGA(string aString)
        {
            return
                 ((IsCorrectCheckSum(aString)) && (Misc.IsGGAMessage(aString)));
        }
        public override bool IsValid()
        {
            if (m_PosFixIndicator==0) return false;

            return true;
        }
        
    }

    public sealed class GLL :   NMEA_Msg
    {
        
        private double m_Latitude;
        private string m_LatitudeStr
        {
            get
            {
                return m_Latitude.ToString();
            }
            set
            {
                MyConvert.TryParseDouble(value, out m_Latitude);
            }
        }
        public  double LatitudeDeg
        {

            get
            {

                return Misc.NmeaFormatDegreesToDegrees(m_Latitude, m_NS_Indicator);
            }
        }

        private string m_NS_Indicator;
        private string m_NS_IndicatorStr
        {
            get { return m_NS_Indicator; }
            set { m_NS_Indicator = value; }
        }

        private double m_Longitude;
        private string m_LongitudeStr
        {
            get
            {
                return m_Longitude.ToString();
            }
            set
            {
                MyConvert.TryParseDouble(value, out m_Longitude);
            }
        }
        public double  LongitudeDeg
        {

            get
            {
                return Misc.NmeaFormatDegreesToDegrees(m_Longitude, m_NS_Indicator);
            }
        }

        private string m_EW_Indicator;
        private string m_EW_IndicatorStr
        {
            get { return m_EW_Indicator; }
            set { m_EW_Indicator = value; }
        }

        private double m_UtcTime;
        private string m_UtcTimeStr
        {
            get
            {
                return m_UtcTime.ToString();
            }
            set
            {
                MyConvert.TryParseDouble(value, out m_UtcTime);
            }
        }

        private string m_Valid;
        private string m_ValidStr
        {
            get
            {
                return m_Valid;
            }
            set
            {
                m_Valid=value;
            }
        }

        private string m_Mode;
        private string m_ModeStr
        {
            get { return m_Mode;  }
            set { m_Mode = value; }
        }

        public GLL(int aFieldsCount)
            : base(aFieldsCount)
        {
        }
        public override string this[int index]
        {
            get
            {
                switch (index)
                {
                    case 0: return m_NameStr;
                    case 1: return m_LatitudeStr;
                    case 2: return m_NS_IndicatorStr;
                    case 3: return m_LongitudeStr;
                    case 4: return m_EW_IndicatorStr;
                    case 5: return m_UtcTimeStr;
                    case 6: return m_ValidStr;
                    case 7: return m_ModeStr;
                    case 8: return m_CSStr;
                    default: return "";
                }
            }
            set
            {
                switch (index)
                {
                    case 0: m_NameStr = value; break;
                    case 1: m_LatitudeStr = value; break;
                    case 2: m_NS_IndicatorStr = value; break;
                    case 3: m_LongitudeStr = value; break;
                    case 4: m_EW_IndicatorStr = value; break;
                    case 5: m_UtcTimeStr = value; break;
                    case 6: m_ValidStr = value; break;
                    case 7: m_ModeStr = value; break;
                    case 8: m_CSStr = value; break;
                }
            }
        }
        public static GLL Parse(string aMessage)
        {
            try
            {
                GLL msg = new GLL(9);
                msg.m_RawMessage = aMessage;
                string[] words = Strings.Split(msg.m_RawMessage, new char[] { ',', '*' }, false);
                for (int i = 0; i < words.Length; i++)
                {
                    if (words[i] == null) continue;
                    msg[i] = words[i];
                }
                return msg;
            }
            catch
            {
                return null;
            }
        }
        public static bool IsCorrectGLL(string aString)
        {
            return
                 ((IsCorrectCheckSum(aString)) && (Misc.IsGLLMessage(aString)));
        }
        public override bool IsValid()
        {
            if (m_Valid != Misc.VALID) return false;

            return true;
        }
    }

    public sealed class GSA :   NMEA_Msg
    {

        private string m_Smode;
        private string m_SmodeStr
        {
            get { return m_Smode;     }
            set { m_Smode = value; }
        }

        private UInt16 m_FixStatus;
        private string m_FixStatusStr
        {
            get
            {
                return m_FixStatus.ToString();
            }
            set
            {
                MyConvert.TryParseUInt16(value, out m_FixStatus);
            }
        }

        private UInt16[] m_SatUsedInChannel = new UInt16[12];
        private string m_SatUsedInChannelStr
        {
            get
            {
                StringBuilder sb = new StringBuilder();
                for (int i = 0 ;i<12;i++)
                {
                    if (m_SatUsedInChannel[i]>0)
                        sb.Append(m_SatUsedInChannel[i].ToString());
                    if (i<11) sb.Append(',');
                }
                return sb.ToString();
            }
            set
            {
                string[] words = Strings.Split(value,new char[] {','},false);
                for (int i = 0; i < words.Length; i++)
                {
                    MyConvert.TryParseUInt16(words[i], out m_SatUsedInChannel[i]);
                }
            }
        }

        private double m_PDOP;
        private string m_PDOPStr
        {
            get { return m_PDOP.ToString(); }
            set { MyConvert.TryParseDouble(value, out m_PDOP); }
        }
        
        
        private double m_HDOP;
        private string m_HDOPStr
        {
            get { return m_HDOP.ToString(); }
            set { MyConvert.TryParseDouble(value, out m_HDOP); }
        }

        private double m_VDOP;
        private string m_VDOPStr
        {
            get { return m_VDOP.ToString(); }
            set { MyConvert.TryParseDouble(value,out m_VDOP); }
        }

        public GSA(int aFieldsCount):base(aFieldsCount)
        {
            
        }
        public override string this[int index]
        {
            get
            {
                switch (index)
                {
                    case 0: return m_NameStr;
                    case 1: return m_SmodeStr;
                    case 2: return m_FixStatusStr;
                    case 3: return m_SatUsedInChannelStr;
                    case 4: return m_PDOPStr ;
                    case 5: return m_HDOPStr;
                    case 6: return m_VDOPStr;
                    case 7: return m_CSStr;
                    default: return "";
                }
            }
            set
            {
                switch (index)
                {
                    case 0: m_NameStr       = value; break;
                    case 1: m_SmodeStr      = value; break;
                    case 2: m_FixStatusStr  = value; break;
                    case 3: m_SatUsedInChannelStr = value; break;
                    case 4: m_PDOPStr       = value; break;
                    case 5: m_HDOPStr       = value; break;
                    case 6: m_VDOPStr       = value; break;
                    case 7: m_CSStr         = value; break;
                }
            }
        }
        public static bool IsCorrectGSA(string aString)
        {
            return
                 ((IsCorrectCheckSum(aString)) && (Misc.IsGSAMessage(aString)));
        }
        public static GSA Parse(string aMessage)
        {
            try
            {
                GSA msg = new GSA(8);
                msg.m_RawMessage = aMessage;
                string[] words = Strings.Split(msg.m_RawMessage, new char[] { ',','*'},false);
                StringBuilder sb = new StringBuilder();
                int j=0;
                for (int i = 0; i < 3;i++ )
                {
                    if (i >= words.Length) break;
                    msg[i] = words[i];
                }
                j=3;
                for (int i = j; i < words.Length-4; i++)
                    {
                        if (i >= words.Length) break;    
                        sb.Append(words[i]);
                        if (i < words.Length - 5)
                        sb.Append(',');
                    }   
                msg[j] = sb.ToString();
                j++;
                for (int i = words.Length - 4; i < words.Length; i++)
                {
                    if (i > words.Length-1) break;
                    msg[j] = words[i];
                    j++;
                }
                return msg;
            }
            catch
            {
                return null;
            }
        }

        public override bool IsValid()
        {
            //TODO

            return true;
        }


    }

    public sealed class GSV :   NMEA_Msg
    {
        internal class ChannelData
        {
            const int FieldsCount = 4;            

            public bool IsRead = false;

            private UInt16 m_SV_ID;
            private string m_SV_IDStr
            {
                get { return m_SV_ID.ToString(); }
                set { MyConvert.TryParseUInt16(value, out m_SV_ID); }
            }

            private double m_ElvDegree;
            private string m_ElvDegreeStr
            {
                get { return m_ElvDegree.ToString(); }
                set { MyConvert.TryParseDouble(value, out m_ElvDegree); }
            }

            private double m_AzDegree;
            private string m_AzDegreeStr
            {
                get { return m_AzDegree.ToString(); }
                set { MyConvert.TryParseDouble(value, out m_AzDegree); }
            }

            private double m_SNR;
            private string m_SNRStr
            {
                get { return m_SNR.ToString(); }
                set { MyConvert.TryParseDouble(value, out m_SNR); }
            }

            public string this[int index]
            {
                get
                {
                    switch (index)
                    {
                        case 0: return m_SV_IDStr;
                        case 1: return m_ElvDegreeStr;
                        case 2: return m_AzDegreeStr;
                        case 3: return m_SNRStr;
                        default: return "";
                    }
                }
                set
                {
                    switch (index)
                    {
                        case 0: m_SV_IDStr = value; break;
                        case 1: m_ElvDegreeStr = value; break;
                        case 2: m_AzDegreeStr = value; break;
                        case 3: m_SNRStr = value; break;
                    }

                }
            }
            public static ChannelData Parse(string aStr)
            {
                try
                {
                    ChannelData msg = new ChannelData();
                    string[] words = Strings.Split(aStr, new char[] {','}, false);
                    for (int i = 0; i < words.Length; i++)
                    {
                        if (words[i] == null) continue;
                        msg[i] = words[i];
                    }
                    msg.IsRead = true;
                    return msg;
                }
                catch
                {
                    return null;
                }
            }
            public override string ToString()
            {
                StringBuilder sb = new StringBuilder();
                for (int i = 0; i < FieldsCount; i++)
                {
                    if (this[i] != "0") 
                        sb.Append(this[i].ToString());
                    if (i < FieldsCount-1) sb.Append(',');
                }
                return sb.ToString();
            }

        }

        private UInt16 m_NoMsg;
        private string m_NoMsgStr 
        {
            get { return m_NoMsg.ToString(); }
            set { MyConvert.TryParseUInt16(value,out m_NoMsg);}
        }

        private UInt16 m_MsgNo;
        private string m_MsgNoStr 
        {
            get { return m_MsgNo.ToString(); }
            set { MyConvert.TryParseUInt16(value, out m_MsgNo); }
        }

        private UInt16 m_NoSv;
        private string m_NoSvStr 
        {
            get { return m_NoSv.ToString(); }
            set { MyConvert.TryParseUInt16(value,out m_NoSv);}
        }

        private ChannelData[] m_Channels = new ChannelData[4];
        private string m_ChannelsStr
        {
            get 
            {
                StringBuilder sb = new StringBuilder();
                for (int i = 0; i < m_Channels.Length; i++)
                {
                    if (m_Channels[i] == null) break;
                    if (!m_Channels[i].IsRead) continue;
                    sb.Append(m_Channels[i].ToString());
                    sb.Append(',');
                }
                sb.Remove(sb.Length - 1, 1);
                return sb.ToString();
            }
            set
            {
                string[] words = Strings.Split(value,new char[]{','},false);
                for (int i = 0; i < m_Channels.Length; i++)
                {
                    StringBuilder sb = new StringBuilder();
                    int cnt = 0;
                    for (int j = i*4; j < i*4+4;j++ )
                    {
                        if (j > words.Length - 1) break;
                        sb.Append(words[j]);
                        cnt++;
                        if (cnt == 4)
                        {
                            break;
                        } else
                        sb.Append(',');
                    }
                    m_Channels[i] = ChannelData.Parse(sb.ToString()); 
                    
                }
            }
        }
        

        public GSV(int aFieldsCount)
            : base(aFieldsCount)
        {
        }
        
        public override string this[int index]
        {
            get
            {
                switch (index)
                {
                    case 0: return m_NameStr;
                    case 1: return m_NoMsgStr;
                    case 2: return m_MsgNoStr;
                    case 3: return m_NoSvStr;
                    case 4: return m_ChannelsStr;
                    case 5: return m_CSStr;
                    default: return "";
                }
            }
            set
            {
                switch (index)
                {
                    case 0: m_NameStr = value; break;
                    case 1: m_NoMsgStr = value; break;
                    case 2: m_MsgNoStr = value; break;
                    case 3: m_NoSvStr = value; break;
                    case 4: m_ChannelsStr = value; break;
                    case 5: m_CSStr = value; break;
                   
                }
                
            }
        }
        
        public static GSV Parse(string aMessage)
        {
            try
            {
                GSV msg = new GSV(6);
                msg.m_RawMessage = aMessage;
                string[] words = Strings.Split(msg.m_RawMessage, new char[] { ',', '*' }, false);
                StringBuilder sb = new StringBuilder();
                int j = 0;
                for (int i = 0; i < 4; i++)
                {
                    if (i >= words.Length) break;
                    msg[i] = words[i];
                }
                j = 4;
                for (int i = j; i < words.Length-1; i++)
                    {
                        sb.Append(words[i]);
                        if (i < words.Length - 2)
                          sb.Append(',');
                        continue;
                     }

                msg[j] = sb.ToString();
                j++;
                msg[j] = words[words.Length-1];
                return msg;
            }
            catch
            {
                return null;
            }
        }
        public static bool IsCorrectGSV(string aString)
        {
            return
                 ((IsCorrectCheckSum(aString)) && (Misc.IsGSVMessage(aString)));
        }
        public override bool IsValid()
        {
            //TODO

            return true;
        }

    }

    public sealed class RMC :   NMEA_Msg
    {
        
        private double m_UtcTime;
        private string m_UtcTimeStr
        {
            get
            {
                return m_UtcTime.ToString();
            }
            set
            {
                MyConvert.TryParseDouble(value, out m_UtcTime);
            }
        }

        private string m_Status;
        private string m_StatusStr
        {
            get { return m_Status; }
            set { m_Status = value; }
        }

        private double m_Latitude;
        private string m_LatitudeStr
        {
            get
            {
                return m_Latitude.ToString();
            }
            set
            {
                MyConvert.TryParseDouble(value, out m_Latitude);
            }
        }
        public double LatitudeDeg
        {

            get
            {

                return Misc.NmeaFormatDegreesToDegrees(m_Latitude, m_NS_Indicator);
            }
        }

        private string m_NS_Indicator;
        private string m_NS_IndicatorStr
        {
            get { return m_NS_Indicator; }
            set { m_NS_Indicator = value; }
        }

        private double m_Longitude;
        private string m_LongitudeStr
        {
            get
            {
                return m_Longitude.ToString();
            }
            set
            {
                MyConvert.TryParseDouble(value, out m_Longitude);
            }
        }
        public double LongitudeDeg
        {

            get
            {
                return Misc.NmeaFormatDegreesToDegrees(m_Longitude, m_NS_Indicator);
            }
        }

        private string m_EW_Indicator;
        private string m_EW_IndicatorStr
        {
            get { return m_EW_Indicator; }
            set { m_EW_Indicator = value; }
        }

        private double m_SpeedInKnots;
        private string m_SpeedInKnotsStr
        {
            get
            {
                return m_SpeedInKnots.ToString();
            }
            set
            {
                MyConvert.TryParseDouble(value, out m_SpeedInKnots);
            }
        }

        private double m_CourseOverGround;
        private string m_CourseOverGroundStr
        {
            get
            {
                return m_CourseOverGround.ToString();
            }
            set
            {
                MyConvert.TryParseDouble(value, out m_CourseOverGround);
            }
        }

        private DateTime m_Date;
        private string m_DateStr
        {

            get { return m_Date.ToString("ddmmyy"); }
            set { m_Date = DateTime.ParseExact(value,"ddmmyy",null); }
        }

        private double m_MV;
        private string m_MVStr
        {
            get
            {
                return m_MV.ToString();
            }
            set
            {
                MyConvert.TryParseDouble(value, out m_MV);
            }
        }

        private string m_MVE;
        private string m_MVEStr
        {
            get { return m_MVE; }
            set { m_MVE = value; }
        }

        private string m_Mode;
        private string m_ModeStr
        {
            get { return m_Mode; }
            set { m_Mode = value; }
        }

        public RMC(int aFieldsCount):base(aFieldsCount)
        {
        }
        public override string this[int index]
        {
            get
            {
                switch (index)
                {
                    case 0: return m_NameStr;
                    case 1: return m_UtcTimeStr;
                    case 2: return m_StatusStr;
                    case 3: return m_LatitudeStr;
                    case 4: return m_NS_IndicatorStr;
                    case 5: return m_LongitudeStr;
                    case 6: return m_EW_IndicatorStr;
                    case 7: return m_SpeedInKnotsStr;
                    case 8: return m_CourseOverGroundStr;
                    case 9: return m_DateStr;
                    case 10: return m_MVStr;
                    case 11: return m_MVEStr;
                    case 12: return m_ModeStr;
                    case 13: return m_CSStr;
                    default: return "";
                }
            }
            set
            {
                switch (index)
                {
                    case 0: m_NameStr = value; break;
                    case 1: m_UtcTimeStr = value; break;
                    case 2: m_StatusStr = value; break;
                    case 3: m_LatitudeStr = value; break;
                    case 4: m_NS_IndicatorStr = value; break;
                    case 5: m_LongitudeStr = value; break;
                    case 6: m_EW_IndicatorStr = value; break;
                    case 7: m_SpeedInKnotsStr= value; break;
                    case 8: m_CourseOverGroundStr = value; break;
                    case 9: m_DateStr = value; break;
                    case 10: m_MVStr = value; break;
                    case 11: m_MVEStr = value; break;
                    case 12: m_ModeStr = value; break;
                    case 13: m_CSStr   = value; break;
                }
            }
        }
        public static RMC Parse(string aMessage)
        {
            try
            {
                RMC msg = new RMC(14);
                msg.m_RawMessage = aMessage;
                string[] words = Strings.Split(msg.m_RawMessage, new char[] { ',', '*' }, false);
                for (int i = 0; i < words.Length; i++)
                {
                    if (words[i] == null) continue;
                    msg[i] = words[i];
                }
                return msg;
            }
            catch
            {
                return null;
            }
        }
        public static bool IsCorrectRMC(string aString)
        {
            return
                 ((IsCorrectCheckSum(aString)) && (Misc.IsRMCMessage(aString)));
        }
        public override bool IsValid()
        {
            if (m_Status != Misc.VALID) return false;
            return true;
        }
    }

    public sealed class VTG :   NMEA_Msg
    {
        private double m_CourseOG;
        private string m_CourseOGStr
        {
            get
            {
                return m_CourseOG.ToString();
            }
            set
            {
                MyConvert.TryParseDouble(value, out m_CourseOG);
            }
        }

        private string m_T;
        private string m_TStr
        {
            get { return m_T; }
            set { m_T = value; }
        }

        private string m_CourseOGM;
        private string m_CourseOGMStr
        {
            get { return m_CourseOGM; }
            set { m_CourseOGM = value; }
        }

        private string m_M;
        private string m_MStr
        {
            get { return m_M;  }
            set { m_M = value; }
        }

        private double m_SpeedOG;
        private string m_SpeedOGStr
        {
            get
            {
                return m_SpeedOG.ToString();
            }
            set
            {
                MyConvert.TryParseDouble(value, out m_SpeedOG);
            }
        }

        private string m_N;
        private string m_NStr
        {
            get { return m_N; }
            set { m_N = value; }
        }

        private double m_SpeedInKmH;
        private string m_SpeedInKmHStr
        {
            get
            {
                return m_SpeedInKmH.ToString();
            }
            set
            {
                MyConvert.TryParseDouble(value, out m_SpeedInKmH);
            }
        }
       
        private string m_K;
        private string m_KStr
        {
            get { return m_K; }
            set { m_K = value; }
        }
        

        public VTG(int aFieldsCount)
            : base(aFieldsCount)
        {
        }

        public override string this[int index]
        {
            get 
            {
                switch (index)
                {
                    case 0: return m_NameStr;                        
                    case 1: return m_CourseOGStr;
                    case 2: return m_TStr;
                    case 3: return m_CourseOGMStr;
                    case 4: return m_MStr;
                    case 5: return m_SpeedOGStr;
                    case 6: return m_NStr;
                    case 7: return m_SpeedInKmHStr;
                    case 8: return m_KStr;
                    case 9: return m_CSStr;
                    default: return "";
                }
            }
            set
            {
                switch (index)
                {
                    case 0:  m_NameStr = value;break;                        
                    case 1:  m_CourseOGStr= value;break;
                    case 2:  m_TStr= value;break;
                    case 3:  m_CourseOGMStr= value;break;
                    case 4:  m_MStr= value;break;
                    case 5:  m_SpeedOGStr= value;break;
                    case 6:  m_NStr= value;break;
                    case 7:  m_SpeedInKmHStr= value;break;
                    case 8:  m_KStr= value;break;
                    case 9: m_CSStr= value;break;
                }
            }
        }
        public static VTG Parse(string aMessage)
        {
            try
            {
                VTG msg = new VTG(10);
                msg.m_RawMessage = aMessage;
                string[] words = Strings.Split(msg.m_RawMessage, new char[] { ',', '*' }, false);
                for (int i = 0; i < words.Length; i++)
                {
                    if (words[i] == null) continue;
                    msg[i] = words[i];
                }
                return msg;
            }
            catch
            {
                return null;
            }
        }
        public static bool IsCorrectVTG(string aString)
        {
            return
                 ((IsCorrectCheckSum(aString)) && (Misc.IsVTGMessage(aString)));
        }

        public override bool IsValid()
        {
            //TODO

            return true;
        }
    }
}