using System;
using System.Text;
using MyUtils;

namespace MyNMEA {

internal enum NP_STATE
{
    NP_STATE_SOM = 0,		            // Search for start of message
    NP_STATE_CMD = 1,						// Get command
    NP_STATE_DATA = 2,						// Get data
    NP_STATE_CHECKSUM_1 =3,				// Get first checksum character
    NP_STATE_CHECKSUM_2 =4,				// get second checksum character
};

internal class Consts {

    public const sbyte MAXFIELD = 25;
 public const sbyte NP_MAX_CMD_LEN		=	8	;	// maximum command length (NMEA address)
 public const short NP_MAX_DATA_LEN		=	256 ;	// maximum data length
 public const sbyte NP_MAX_CHAN			=	36	;	// maximum number of channels
 public const sbyte NP_WAYPOINT_ID_LEN  =   32  ;	// waypoint max string len
 
 public static readonly string c_Eeast = "E";
 public static readonly string c_West  = "W";
 public static readonly string c_North = "N";
 public static readonly string c_South = "S";
}

public struct CNPSatInfo
{
    public  int m_wPRN;						//
    public int m_wSignalQuality;			//
    public bool m_bUsedInSolution;			//
    public int m_wAzimuth;					//
    public int m_wElevation;				//
};


 public  class CNMEAParser
    {
        
        private bool m_ValidMsg = false;
        private NMEA m_LastValidMsg;
        private NP_STATE m_nState;					// Current state protocol parser is in
        private byte m_btChecksum;					// Calculated NMEA sentence checksum
        private byte m_btReceivedChecksum;			// Received NMEA sentence checksum (if exists)
        private int m_wIndex;						// Index used for command and data
        private byte[] m_pCommand = new byte[Consts.NP_MAX_CMD_LEN];	// NMEA command
        private byte[] m_pData = new byte[Consts.NP_MAX_DATA_LEN];		// NMEA data

        private StringBuilder sbRawRMC = new StringBuilder();
        private string GetRawMsg()
        {
            return
                 string.Format("${0},{1}*{2}",Strings.MakeNullTerminatedString(m_pCommand),Strings.MakeNullTerminatedString(m_pData),m_btReceivedChecksum.ToString("X2"));
        }

     public string RawRMC
     {
         get
         {
             return sbRawRMC.ToString();
         }
     }
        public int m_dwCommandCount;				// number of NMEA commands received (processed or not processed)
        public NMEA LastValidMsg
        {
         get { return m_LastValidMsg; }
        }
        //
        // GPGGA Data
        //
        public byte m_btGGAHour;					//
        public byte m_btGGAMinute;					//
        public byte m_btGGASecond;					//
        public double m_dGGALatitude;				// < 0 = South, > 0 = North
        public double m_dGGALongitude;				// < 0 = West, > 0 = East
        public byte m_btGGAGPSQuality;				// 0 = fix not available, 1 = GPS sps mode, 2 = Differential GPS, SPS mode, fix valid, 3 = GPS PPS mode, fix valid
        public byte m_btGGANumOfSatsInUse;			//
        public double m_dGGAHDOP;					//
        public double m_dGGAAltitude;				// Altitude: mean-sea-level (geoid) meters
        public int m_dwGGACount;					//
        public int m_nGGAOldVSpeedSeconds;			//
        public double m_dGGAOldVSpeedAlt;			//
        public double m_dGGAVertSpeed;				//   
        
        //
	// GPGSA
	//
        public byte m_btGSAMode;					// M = manual, A = automatic 2D/3D
        public byte m_btGSAFixMode;				// 1 = fix not available, 2 = 2D, 3 = 3D
        public int[] m_wGSASatsInSolution = new int[Consts.NP_MAX_CHAN]; // ID of sats in solution
        public double m_dGSAPDOP;					//
        public double m_dGSAHDOP;					//
        public double m_dGSAVDOP;					//
        public int m_dwGSACount;					//

	//
	// GPGSV
	//
        public  byte m_btGSVTotalNumOfMsg;			//
        public int m_wGSVTotalNumSatsInView;		//
        public CNPSatInfo[] m_GSVSatInfo = new CNPSatInfo[Consts.NP_MAX_CHAN];	//
        public int m_dwGSVCount;					//

        //
        // GPRMC
        //
        public byte m_btRMCHour;					//
        public byte m_btRMCMinute;					//
        public byte m_btRMCSecond;					//
        public byte m_btRMCDataValid;				// A = Data valid, V = navigation rx warning
        public double m_dRMCLatitude;				// current latitude
        public double m_dRMCLongitude;				// current longitude
        public double m_dRMCGroundSpeed;			// speed over ground, knots
        public double m_dRMCCourse;				// course over ground, degrees true
        byte m_btRMCDay;					//
        public byte m_btRMCMonth;					//
        public int m_wRMCYear;					//
        public double m_dRMCMagVar;				// magnitic variation, degrees East(+)/West(-)
        public int m_dwRMCCount;					//

        private void ProcessGPGGA(byte[] pData)
        {
        	byte[] pField = new byte[Consts.MAXFIELD];
	        char[] pBuff = new char[10];
	//
	// Time
	//
	if(GetField(pData, pField, 0, Consts.MAXFIELD))
	{
		// Hour
		pBuff[0] = Convert.ToChar(pField[0]);
		pBuff[1] = Convert.ToChar(pField[1]);
		pBuff[2] = '\0';
		m_btGGAHour = byte.Parse(Strings.MakeNullTerminatedString(pBuff));

        // minute
		pBuff[0] = Convert.ToChar(pField[2]);
		pBuff[1] = Convert.ToChar(pField[3]);
		pBuff[2] = '\0';
		m_btGGAMinute = byte.Parse(Strings.MakeNullTerminatedString(pBuff));

		// Second
		pBuff[0] = Convert.ToChar(pField[4]);
		pBuff[1] = Convert.ToChar(pField[5]);
		pBuff[2] = '\0';
		m_btGGASecond = byte.Parse(Strings.MakeNullTerminatedString(pBuff));
	}

	//
	// Latitude
	//
	if(GetField(pData, pField, 1, Consts.MAXFIELD))
	{
		string fld = Strings.MakeNullTerminatedString(pField);
        m_dGGALatitude = double.Parse(fld);
        m_dGGALatitude = Misc.NmeaFormatDegreesToDegrees(m_dGGALatitude);
	}
	if(GetField(pData, pField, 2, Consts.MAXFIELD))
	{
		if(pField[0] == 'S')
		{
			m_dGGALatitude = -m_dGGALatitude;
		}
	}

	//
	// Longitude
	//
	if(GetField(pData, pField, 3, Consts.MAXFIELD))
	{
		string fld = Strings.MakeNullTerminatedString(pField);
        m_dGGALongitude  = double.Parse(fld);
        m_dGGALongitude = Misc.NmeaFormatDegreesToDegrees(m_dGGALongitude);
	}
	if(GetField(pData, pField, 4, Consts.MAXFIELD))
	{
		if(pField[0] == 'W')
		{
			m_dGGALongitude = -m_dGGALongitude;
		}
	}

	//
	// GPS quality
	//
	if(GetField(pData, pField, 5, Consts.MAXFIELD))
	{
		m_btGGAGPSQuality = (byte)(pField[0] - Convert.ToByte('0'));
	}

	//
	// Satellites in use
	//
	if(GetField(pData, pField, 6, Consts.MAXFIELD))
	{
		pBuff[0] = Convert.ToChar(pField[0]);
		pBuff[1] = Convert.ToChar(pField[1]);
		pBuff[2] = '\0';
		m_btGGANumOfSatsInUse = byte.Parse(Strings.MakeNullTerminatedString(pBuff) );
	}

	//
	// HDOP
	//
	if(GetField(pData, pField, 7, Consts.MAXFIELD))
	{
		m_dGGAHDOP =  double.Parse(Strings.MakeNullTerminatedString(pField));
	}
	
	//
	// Altitude
	//
    if (GetField(pData, pField, 8, Consts.MAXFIELD))
	{
        m_dGGAAltitude = double.Parse(Strings.MakeNullTerminatedString(pField));
	}

	//
	// Durive vertical speed (bonus)
	//
	int nSeconds = (int)m_btGGAMinute * 60 + (int)m_btGGASecond;
	if(nSeconds > m_nGGAOldVSpeedSeconds)
	{
		double dDiff = (double)(m_nGGAOldVSpeedSeconds-nSeconds);
		double dVal = dDiff/60.0;
		if(dVal != 0.0)
		{
			m_dGGAVertSpeed = (m_dGGAOldVSpeedAlt - m_dGGAAltitude) / dVal;
		}
	}
	m_dGGAOldVSpeedAlt = m_dGGAAltitude;
	m_nGGAOldVSpeedSeconds = nSeconds;
    
            m_dwGGACount++;
        }
        private void ProcessGPGSA(byte[] pData)
        {
            	byte[] pField = new byte[Consts.MAXFIELD];
	            char[] pBuff  = new char[10];

	//
	// Mode
	//
                if (GetField(pData,pField,0,Consts.MAXFIELD))
	            {
		            m_btGSAMode = pField[0];
	            }

	//
	// Fix Mode
	//
	if(GetField(pData, pField, 1, Consts.MAXFIELD))
	{
		m_btGSAFixMode = (byte)(pField[0] - Convert.ToByte('0'));
	}

	//
	// Active satellites
	//
	for(int i = 0; i < 12; i++)
	{
		if(GetField(pData, pField, 2 + i, Consts.MAXFIELD))
		{
			pBuff[0] = Convert.ToChar(pField[0]);
			pBuff[1] = Convert.ToChar(pField[1]);
			pBuff[2] = '\0';
			m_wGSASatsInSolution[i] = int.Parse(Strings.MakeNullTerminatedString(pBuff));
		}
		else
		{
			m_wGSASatsInSolution[i] = 0;
		}
	}

	//
	// PDOP
	//
	if(GetField(pData, pField, 14, Consts.MAXFIELD))
	{
		m_dGSAPDOP = double.Parse(Strings.MakeNullTerminatedString(pField));
	}
	else
	{
		m_dGSAPDOP = 0.0;
	}

	//
	// HDOP
	//
	if(GetField(pData, pField, 15, Consts.MAXFIELD))
	{
		m_dGSAHDOP = double.Parse(Strings.MakeNullTerminatedString(pField));
	}
	else
	{
		m_dGSAHDOP = 0.0;
	}

	//
	// VDOP
	//
	if(GetField(pData, pField, 16, Consts.MAXFIELD))
	{
		m_dGSAVDOP = double.Parse(Strings.MakeNullTerminatedString(pField));
	}
	else
	{
		m_dGSAVDOP = 0.0;
	}

	m_dwGSACount++;
        }
        private void ProcessGPRMC(byte[] pData)
        {
            char[] pBuff  = new char [10];
	        byte[] pField = new byte [Consts.MAXFIELD];
            //string strField = null;
	//
	// Time
	//
            if (GetField(pData, pField, 0, Consts.MAXFIELD))
	{
		// Hour
		pBuff[0] = Convert.ToChar(pField[0]);
		pBuff[1] = Convert.ToChar(pField[1]);
        pBuff[2] ='\0';
        m_btRMCHour = byte.Parse(Strings.MakeNullTerminatedString(pBuff));

		// minute
		pBuff[0] = Convert.ToChar(pField[2]);
		pBuff[1] = Convert.ToChar(pField[3]);
		pBuff[2] = '\0';
        m_btRMCMinute = byte.Parse(Strings.MakeNullTerminatedString(pBuff));

		// Second
		pBuff[0] = Convert.ToChar(pField[4]);
		pBuff[1] = Convert.ToChar(pField[5]);
		pBuff[2] = '\0';
        m_btRMCSecond = byte.Parse(Strings.MakeNullTerminatedString(pBuff));
	}

	//
	// Data valid
	//
	if(GetField(pData, pField, 1, Consts.MAXFIELD))
	{
		m_btRMCDataValid = pField[0];
	}
	else
	{
		m_btRMCDataValid = Convert.ToByte('V');
	}

	//
	// latitude
	//
	if(GetField(pData,pField,2,Consts.MAXFIELD))
	{
        string fld = Strings.MakeNullTerminatedString(pField);
        m_dRMCLatitude = double.Parse(fld);
        m_dRMCLatitude = Misc.NmeaFormatDegreesToDegrees(m_dRMCLatitude);
	}
	if(GetField(pData, pField, 3, Consts.MAXFIELD))
	{
		if(pField[0] == Convert.ToByte('S'))
		{
			m_dRMCLatitude = -m_dRMCLatitude;
		}
	}

	//
	// Longitude
	//
	if(GetField(pData, pField, 4, Consts.MAXFIELD))
	{
        string fld = Strings.MakeNullTerminatedString(pField);
        m_dRMCLongitude = double.Parse(fld);
        m_dRMCLongitude = Misc.NmeaFormatDegreesToDegrees(m_dRMCLongitude);
	}

    if (GetField(pData, pField, 5,Consts.MAXFIELD))
	{
		if(pField[0] == 'W')
		{
			m_dRMCLongitude = -m_dRMCLongitude;
		}
	}

	//
	// Ground speed
	//
    if (GetField(pData, pField, 6, Consts.MAXFIELD))
	{
         string str = Strings.MakeNullTerminatedString(pField);
         m_dRMCGroundSpeed = double.Parse(str);
	}
	else
	{
		m_dRMCGroundSpeed = 0.0;
	}

	//
	// course over ground, degrees true
	//
    if (GetField(pData, pField, 7, Consts.MAXFIELD))
	{
        m_dRMCCourse = double.Parse(Strings.MakeNullTerminatedString(pField));
	}
	else
	{
		m_dRMCCourse = 0.0;
	}

	//
	// Date
	//
	if(GetField(pData, pField, 8, Consts.MAXFIELD))
	{
		// Day
		pBuff[0] = Convert.ToChar(pField[0]);
		pBuff[1] = Convert.ToChar(pField[1]);
		pBuff[2] = '\0';
        m_btRMCDay = byte.Parse(Strings.MakeNullTerminatedString(pBuff));

		// Month
		pBuff[0] = Convert.ToChar(pField[2]);
		pBuff[1] = Convert.ToChar(pField[3]);
		pBuff[2] = '\0';
        m_btRMCMonth = byte.Parse(Strings.MakeNullTerminatedString(pBuff));

		// Year (Only two digits. I wonder why?)
		pBuff[0] = Convert.ToChar(pField[4]);
		pBuff[1] = Convert.ToChar(pField[5]);
		pBuff[2] = '\0';
        m_wRMCYear = int.Parse(Strings.MakeNullTerminatedString(pBuff));
		m_wRMCYear += 2000;				// make 4 digit date -- What assumptions should be made here?
	}

	//
	// course over ground, degrees true
	//
	if(GetField(pData, pField, 9, Consts.MAXFIELD))
	{
        m_dRMCMagVar = double.Parse(Strings.MakeNullTerminatedString(pField)); 
	}
	else
	{
		m_dRMCMagVar = 0.0;
	}
	if(GetField(pData, pField, 10, Consts.MAXFIELD))
	{
		if(pField[0] == 'W')
		{
			m_dRMCMagVar = -m_dRMCMagVar;
		}
	}

	m_dwRMCCount++;

  }
        private void ProcessGPGSV(byte[] pData)
        {
            byte nTotalNumOfMsg=0;
            int nMsgNum = 0;
	        byte[] pField = new byte[Consts.MAXFIELD];

	//
	// Total number of messages
	//
	if(GetField(pData, pField, 0, Consts.MAXFIELD))
	{
		nTotalNumOfMsg = byte.Parse(Strings.MakeNullTerminatedString(pField));

		//
		// Make sure that the nTotalNumOfMsg is valid. This is used to
		// calculate indexes into an array. I've seen corrept NMEA strings
		// with no checksum set this to large values.
		//
		if(nTotalNumOfMsg > 9 || nTotalNumOfMsg < 0) return; 
	}
	if(nTotalNumOfMsg < 1 || nTotalNumOfMsg*4 >= Consts.NP_MAX_CHAN)
	{
		return;
	}
    
	//
	// message number
	//
    m_btGSVTotalNumOfMsg = nTotalNumOfMsg;  //?????? NIKI

	if(GetField(pData, pField, 1, Consts.MAXFIELD))
	{
		nMsgNum = int.Parse(Strings.MakeNullTerminatedString(pField));

		//
		// Make sure that the message number is valid. This is used to
		// calculate indexes into an array
		//
		if(nMsgNum > 9 || nMsgNum < 0) return; 
	}

	//
	// Total satellites in view
	//
	if(GetField(pData, pField, 2, Consts.MAXFIELD))
	{
		m_wGSVTotalNumSatsInView = int.Parse(Strings.MakeNullTerminatedString(pField));
	}

	//
	// Satelite data
	//
	for(int i = 0; i < 4; i++)
	{
		// Satellite ID
		if(GetField(pData, pField, 3 + 4*i, Consts.MAXFIELD))
		{
			m_GSVSatInfo[i+(nMsgNum-1)*4].m_wPRN = int.Parse(Strings.MakeNullTerminatedString(pField));
		}
		else
		{
			m_GSVSatInfo[i+(nMsgNum-1)*4].m_wPRN = 0;
		}

		// Elevarion
		if(GetField(pData, pField, 4 + 4*i,Consts.MAXFIELD))
		{
			m_GSVSatInfo[i+(nMsgNum-1)*4].m_wElevation = int.Parse(Strings.MakeNullTerminatedString(pField));
		}
		else
		{
			m_GSVSatInfo[i+(nMsgNum-1)*4].m_wElevation = 0;
		}

		// Azimuth
		if(GetField(pData, pField, 5 + 4*i, Consts.MAXFIELD))
		{
			m_GSVSatInfo[i+(nMsgNum-1)*4].m_wAzimuth = int.Parse(Strings.MakeNullTerminatedString(pField));
		}
		else
		{
			m_GSVSatInfo[i+(nMsgNum-1)*4].m_wAzimuth = 0;
		}

		// SNR
		if(GetField(pData, pField, 6 + 4*i, Consts.MAXFIELD))
		{
			m_GSVSatInfo[i+(nMsgNum-1)*4].m_wSignalQuality = int.Parse(Strings.MakeNullTerminatedString(pField));
		}
		else
		{
			m_GSVSatInfo[i+(nMsgNum-1)*4].m_wSignalQuality = 0;
		}

		//
		// Update "used in solution" (m_bUsedInSolution) flag. This is base
		// on the GSA message and is an added convenience for post processing
		//
		m_GSVSatInfo[i+(nMsgNum-1)*4].m_bUsedInSolution = IsSatUsedInSolution(m_GSVSatInfo[i+(nMsgNum-1)*4].m_wPRN);
	    }
	        m_dwGSVCount++;
        }
        
        private bool IsSatUsedInSolution(int wSatID)
        {
            /*
            if (wSatID == 0) return false;
            for (int i = 0; i < 12; i++)
            {
                if (wSatID == m_wGSASatsInSolution[i])
                {
                    return TRUE;
                }
            }
            */
            return false;
        }
        public  void Reset()
        {
            int i=0;

            //
            // GPGGA Data
            //
            m_btGGAHour = 0;					//
            m_btGGAMinute = 0;					//
            m_btGGASecond = 0;					//
            m_dGGALatitude = 0.0;				// < 0 = South, > 0 = North 
            m_dGGALongitude = 0.0;				// < 0 = West, > 0 = East
            m_btGGAGPSQuality = 0;				// 0 = fix not available, 1 = GPS sps mode, 2 = Differential GPS, SPS mode, fix valid, 3 = GPS PPS mode, fix valid
            m_btGGANumOfSatsInUse = 0;			//
            m_dGGAHDOP = 0.0;					//
            m_dGGAAltitude = 0.0;				// Altitude: mean-sea-level (geoid) meters
            m_dwGGACount = 0;					//
            m_nGGAOldVSpeedSeconds = 0;			//
            m_dGGAOldVSpeedAlt = 0.0;			//
            m_dGGAVertSpeed = 0.0;				//

            //
            // GPGSA
            //
            m_btGSAMode = Convert.ToByte('M');					// M = manual, A = automatic 2D/3D
            m_btGSAFixMode = 1;					// 1 = fix not available, 2 = 2D, 3 = 3D
            for (i = 0; i < Consts.NP_MAX_CHAN; i++)
            {
                m_wGSASatsInSolution[i] = 0;	// ID of sats in solution
            }
            m_dGSAPDOP = 0.0;					//
            m_dGSAHDOP = 0.0;					//
            m_dGSAVDOP = 0.0;					//
            m_dwGSACount = 0;					//

            //
            // GPGSV
            //
            m_btGSVTotalNumOfMsg = 0;			//
            m_wGSVTotalNumSatsInView = 0;		//
            for (i = 0; i < Consts.NP_MAX_CHAN; i++)
            {
                m_GSVSatInfo[i].m_wAzimuth = 0;
                m_GSVSatInfo[i].m_wElevation = 0;
                m_GSVSatInfo[i].m_wPRN = 0;
                m_GSVSatInfo[i].m_wSignalQuality = 0;
                m_GSVSatInfo[i].m_bUsedInSolution = false;
            }
            m_dwGSVCount = 0;

            //
            // GPRMC
            //
            m_btRMCHour = 0;					//
            m_btRMCMinute = 0;					//
            m_btRMCSecond = 0;					//
            m_btRMCDataValid =  Convert.ToByte('V');  	// A = Data valid, V = navigation rx warning
            m_dRMCLatitude = 0.0;				// current latitude
            m_dRMCLongitude = 0.0;				// current longitude
            m_dRMCGroundSpeed = 0.0;			// speed over ground, knots
            m_dRMCCourse = 0.0;					// course over ground, degrees true
            m_btRMCDay = 1;						//
            m_btRMCMonth = 1;					//
            m_wRMCYear   = 2000;					//
            m_dRMCMagVar = 0.0;					// magnitic variation, degrees East(+)/West(-)
            m_dwRMCCount = 0;					//

            m_nState = NP_STATE.NP_STATE_SOM; //NIKI
        }
        private bool GetField(byte[] pData, byte[] pField, int nFieldNum, int nMaxFieldLen)
        {
            //
            // Validate params
            //
            if (pData == null || pField == null || nMaxFieldLen <= 0)
            {
                return false;
            }

            //
            // Go to the beginning of the selected field
            //
            int i = 0;
            int nField = 0;
            while ((nField != nFieldNum) && (pData[i] != Convert.ToByte('\0')))
            {
                if (pData[i] == ',')
                {
                    nField++;
                }
                i++;
                if (pData[i] == Convert.ToByte('\0'))
                {
                    pField[0] = Convert.ToByte('\0');
                    return false;
                }
            }

            if (pData[i] == ',' || pData[i] == '*')
            {
                pField[0] = Convert.ToByte('\0');
                return false;
            }

            //
            // copy field from pData to Field
            //
            int i2 = 0;
            while ((pData[i] != Convert.ToByte('\0') ) && (pData[i] != Convert.ToByte(',')) && (pData[i] != Convert.ToByte('*')))
            {
                pField[i2] = pData[i];
                i2++; i++;

                //
                // check if field is too big to fit on passed parameter. If it is,
                // crop returned field to its max length.
                //
                if (i2 >= nMaxFieldLen)
                {
                    i2 = nMaxFieldLen - 1;
                    break;
                }
            }
            pField[i2] = Convert.ToByte('\0');
           return true;
        }
        private bool ProcessCommand(byte[] pCommand, byte[] pData)
        {
            if (string.Compare(Encoding.ASCII.GetString(pCommand), "GPRMC") == 0)
            {
                ProcessGPRMC(pData);
                sbRawRMC.Remove(0, sbRawRMC.Length);
                sbRawRMC.Append(GetRawMsg());
                m_ValidMsg = ( m_btRMCDataValid == Convert.ToByte('A'));
                if (m_ValidMsg)
                {
                    m_LastValidMsg = NMEA.RMC;
                }
            } else
            if (string.Compare(Encoding.ASCII.GetString(pCommand), "GPGGA") == 0)
            {
                ProcessGPGGA(pData);
                m_ValidMsg = (m_btGGAGPSQuality != 0);
                if (m_ValidMsg)
                {
                    m_LastValidMsg = NMEA.GGA;
                }
            }
            else
            if (string.Compare(Encoding.ASCII.GetString(pCommand), "GPGSA") == 0)
            {
                ProcessGPGSA(pData);
                m_ValidMsg = (m_btGSAFixMode > 1);
                if (m_ValidMsg)
                {
                    m_LastValidMsg = NMEA.GSA;
                }
            }
            else
            if (string.Compare(Encoding.ASCII.GetString(pCommand), "GPGSV") == 0)
            {
                ProcessGPGSV(pData);
                m_ValidMsg = (m_wGSVTotalNumSatsInView>=4);
                if (m_ValidMsg)
                {
                    m_LastValidMsg = NMEA.GSV;
                }
            }
            else 
            {
                m_ValidMsg = false;
                m_LastValidMsg = NMEA.None;
                return false;
            }
            m_dwCommandCount++;
            return true;
        }

        private void ProcessNMEA(byte btData)
        {
            switch (m_nState)
            {
                ///////////////////////////////////////////////////////////////////////
                // Search for start of message '$'
                case NP_STATE.NP_STATE_SOM:
                    if (btData == '$')
                    {
                        m_btChecksum = 0;			// reset checksum
                        m_wIndex = 0;				// reset index
                        m_nState = NP_STATE.NP_STATE_CMD;
                    }
                    break;

                ///////////////////////////////////////////////////////////////////////
                // Retrieve command (NMEA Address)
                case NP_STATE.NP_STATE_CMD:
                    if (btData != ',' && btData != '*')
                    {
                        m_pCommand[m_wIndex++] = btData;
                        m_btChecksum ^= btData;

                        // Check for command overflow
                        if (m_wIndex >= Consts.NP_MAX_CMD_LEN)
                        {
                            m_nState = NP_STATE.NP_STATE_SOM;
                        }
                    }
                    else
                    {
                        m_pCommand[m_wIndex] = Convert.ToByte('\0');	// terminate command
                        m_btChecksum ^= btData;
                        m_wIndex = 0;
                        m_nState = NP_STATE.NP_STATE_DATA;		// goto get data state
                    }
                    break;

                ///////////////////////////////////////////////////////////////////////
                // Store data and check for end of sentence or checksum flag
                case NP_STATE.NP_STATE_DATA:
                    if (btData == '*') // checksum flag?
                    {
                        m_pData[m_wIndex] = Convert.ToByte('\0');
                        m_nState = NP_STATE.NP_STATE_CHECKSUM_1;
                    }
                    else // no checksum flag, store data
                    {
                        //
                        // Check for end of sentence with no checksum
                        //
                        if (btData == '\r')
                        {
                            m_pData[m_wIndex] = Convert.ToByte('\0');
                            ProcessCommand(m_pCommand, m_pData);
                            m_nState = NP_STATE.NP_STATE_SOM;
                            return;
                        }

                        //
                        // Store data and calculate checksum
                        //
                        m_btChecksum ^= btData;
                        m_pData[m_wIndex] = btData;
                        if (++m_wIndex >= Consts.NP_MAX_DATA_LEN) // Check for buffer overflow
                        {
                            m_nState = NP_STATE.NP_STATE_SOM;
                        }
                    }
                    break;

                ///////////////////////////////////////////////////////////////////////
                case NP_STATE.NP_STATE_CHECKSUM_1:
                    if ((btData - '0') <= 9)
                    {
                        m_btReceivedChecksum = (byte)( (btData - Convert.ToByte('0')) << 4);
                    }
                    else
                    {
                        m_btReceivedChecksum = (byte)( (btData - 'A' + 10) << 4 );
                    }

                    m_nState = NP_STATE.NP_STATE_CHECKSUM_2;

                    break;

                ///////////////////////////////////////////////////////////////////////
                case NP_STATE.NP_STATE_CHECKSUM_2:
                    if ((btData - '0') <= 9)
                    {
                        m_btReceivedChecksum |= (byte)(btData - '0');
                    }
                    else
                    {
                        m_btReceivedChecksum |= (byte)(btData - 'A' + 10);
                    }

                    if (m_btChecksum == m_btReceivedChecksum)
                    {
                        ProcessCommand(m_pCommand, m_pData);
                    }

                    m_nState = NP_STATE.NP_STATE_SOM;
                    break;

                ///////////////////////////////////////////////////////////////////////
                default: m_nState = NP_STATE.NP_STATE_SOM; break;
            }
        }
        public bool ParseBuffer(byte[] pBuff, int dwLen)
        {
            for (int i = 0; i < dwLen; i++)
            {
                ProcessNMEA(pBuff[i]);
            }
            return true;
        }
        public CNMEAParser()
        {
            m_nState = NP_STATE.NP_STATE_SOM;
            m_dwCommandCount = 0;
            Reset();
        }
    }
};
