using System;
using System.Text;
using System.Drawing;
using System.Windows.Forms;
using MyUtils.Controls;


/// <summary>
/// Summary description for Class1
/// </summary>
namespace MyUtils
{
    public class Consts 
    {
        public static readonly System.Globalization.NumberFormatInfo nfi = new System.Globalization.CultureInfo("en-US", false).NumberFormat;
        
        public static readonly Size PixelsPerDegree = new Size(1000000, 1000000);
    }
    
    
    public enum MousePointer { None, ZoomIn, ZoomOut, Move, Select,Point,Line,PLine,Region}
    
    public sealed class Strings
    {
        public static bool IsNullOrEmpty(string aString)
        {
            if (aString == null) return true;
            if (aString == "") return true;
            return false;
        }
        
        public static string[] Split(string aString, char[] aDelimiters,bool RemoveEmpty)
        {
            string[] result;
            string[] tmp = aString.Split(aDelimiters);
            result = tmp;
            if (RemoveEmpty)
            {
                int cnt = 0;
                for (int i = 0; i < tmp.Length; i++)
                {
                    if (tmp[i] == string.Empty) continue;
                    cnt++;
                }
                result = new string[cnt];
                int j = 0;
                for (int i = 0; i < tmp.Length; i++)
                {
                    if (tmp[i] == string.Empty) continue;
                    result[j] = tmp[i];
                    j++;
                }
            }
            return result;
        }

        public static string[] Split(string aString, char[] aDelimiters)
        {
            return Split(aString, aDelimiters, true);
        }

        public static string MakeNullTerminatedString(char[] aCharArray)
        {
            string result = new string(aCharArray);
            result = result.Remove(result.IndexOf('\0'),result.Length-result.IndexOf('\0'));
            return result;
        }
        public static string MakeNullTerminatedString(byte[] aByteArray,Encoding aEncoding)
        {
            string result = aEncoding.GetString(aByteArray);
            result = result.Remove(result.IndexOf('\0'), result.Length - result.IndexOf('\0'));
            return result;
        }

        public static string MakeNullTerminatedString(byte[] aByteArray)
        {
            string result = Encoding.ASCII.GetString(aByteArray);
            result = result.Remove(result.IndexOf('\0'), result.Length - result.IndexOf('\0'));
            return result;
        }

        public static void ItemsAdd(ComboBox.ObjectCollection aItems,string aString,string aStrIndex)
        {
            aItems.Add(string.Format("{0}{1,400}",aString,aStrIndex));
        }

        public static void ItemsAdd(ComboBox.ObjectCollection aItems,string aString, string aStrIndex,int aImageIndex)
        {
            aItems.Add(new ComboBoxEx.ComboBoxExItem(string.Format("{0}{1,400}", aString, aStrIndex),aImageIndex));
        }

        public static void ComboAdd(ComboBox aCombo, string aString, string aStrIndex)
        {
            ItemsAdd(aCombo.Items, aString, aStrIndex);
        }

        public static void ComboAdd(ComboBoxEx aCombo, string aString, string aStrIndex,int aImageIndex)
        {
            ItemsAdd(aCombo.Items, aString, aStrIndex,aImageIndex);
        }


        public static string ItemsGet(ComboBox.ObjectCollection aItems, int Idx)
        {
            if (Idx < 0) return "";
            string tmp ="";
            if (aItems[Idx] is ComboBoxEx.ComboBoxExItem)
                tmp = (aItems[Idx] as ComboBoxEx.ComboBoxExItem).Text;
            else
                tmp = aItems[Idx] as string;
            return
                  (tmp.Substring((tmp.Length - 400), 400).Trim());
        }

        public static string ComboGet(ComboBox aCombo)
        {
            return ItemsGet(aCombo.Items, aCombo.SelectedIndex);
        }

        public static void ComboFind(ComboBox aCombo,string aStrIndex)
        {
            for (int i = 0; i < aCombo.Items.Count; i++)
            {
                if (aStrIndex == ItemsGet(aCombo.Items, i))
                {
                    aCombo.SelectedIndex = i;
                    return;
                }
            }
        }



    }

    public struct PointDouble
    {
        public double X;
        public double Y;
        public PointDouble(double aX,double aY)
        {
            X = aX;
            Y = aY;
        }
    }

    public struct SizeDouble
    {
        public double Width;
        public double Height;
        public SizeDouble(double aWidth,double aHeight)
        {
            Width  = aWidth;
            Height = aHeight;
        }
    }

    public sealed class MapTrans
    {
        private static double mAspectRatio;
        private static bool isInit = false;
        private static PointDouble mBasePoint=new PointDouble();
        public  static Size mPixelsPerDegree=new Size();
        private static Size mBitmapSize=new Size();
        private static SizeDouble mMapSize=new SizeDouble();

        public static bool IsInit { get { return isInit; } set { isInit = value; } }
        public  static PointDouble BasePoint {
            get { return mBasePoint; }
            set { mBasePoint = value; }
        }
        public  static Size BitmapSize { get { return mBitmapSize; } }


        public static void GetPixel(double deg_x, double deg_y, out int pix_x, out int pix_y)
        {
            if (!isInit) throw new ApplicationException("MapTrans not initialized!");
            pix_x = (int)((deg_x - mBasePoint.X) * mPixelsPerDegree.Width) + mBitmapSize.Width/2;
            pix_y = mBitmapSize.Height - ((int)((deg_y - mBasePoint.Y) * mPixelsPerDegree.Height) + mBitmapSize.Height / 2);
        }
        public static void GetPixel(PointDouble deg_point, Point out_pix_point)
        {
            int pix_x, pix_y;
            GetPixel(deg_point.X, deg_point.Y, out pix_x ,out pix_y);
            out_pix_point.X = pix_x;
            out_pix_point.Y = pix_y;
        }
        public static Point GetPixel(PointDouble deg_point)
        {
            int pix_x, pix_y;
            GetPixel(deg_point.X, deg_point.Y, out pix_x, out pix_y);
            return new Point(pix_x, pix_y);
        }


      
        public static void GetDegree(int pix_x, int pix_y, out double deg_x, out double deg_y)
        {
            if (!isInit) throw new ApplicationException("MapTrans not initialized!");
            deg_x = (double)(pix_x - mBitmapSize.Width / 2) / mPixelsPerDegree.Width  +  mBasePoint.X;
            deg_y = (double)(mBitmapSize.Height - (pix_y + mBitmapSize.Height / 2)) / mPixelsPerDegree.Height + mBasePoint.Y ;
        }
        public static void GetDegree(Point pix_point, PointDouble out_deg_point)
        {
            double deg_x,deg_y;
            GetDegree(pix_point.X, pix_point.Y,out deg_x,out deg_y);
            out_deg_point.X = deg_x;
            out_deg_point.Y = deg_y;
        }
        public static PointDouble GetDegree(Point pix_point)
        {
            double deg_x,deg_y;
            GetDegree(pix_point.X, pix_point.Y,out deg_x,out deg_y);
            return new PointDouble(deg_x, deg_y);
        }

        
        public static void Init(double aBasePoint_x,double aBasePoint_y,double aMapSizeDeg_width,double aMapSizeDeg_height,int aBitmapSize_width,int aBitmapSize_height)
        {
            try
            {
                mBasePoint.X = aBasePoint_x;
                mBasePoint.Y = aBasePoint_y;
                mMapSize.Width  = aMapSizeDeg_width;
                mMapSize.Height = aMapSizeDeg_height;
                mBitmapSize.Width  = aBitmapSize_width;
                mBitmapSize.Height = aBitmapSize_height;
                if ((aMapSizeDeg_width > 0) && (aMapSizeDeg_height > 0))
                {
                    mPixelsPerDegree.Width = (int)(aBitmapSize_width / aMapSizeDeg_width);
                    mPixelsPerDegree.Height = (int)(aBitmapSize_height / aMapSizeDeg_height);
                }
                else
                {
                    mPixelsPerDegree = Consts.PixelsPerDegree;
                }
                mAspectRatio = (double)mPixelsPerDegree.Height / mPixelsPerDegree.Width;
            }
            catch
            {
                isInit = false;
                return;
            }
            isInit              = true;

        }
        public static void Init(PointDouble aBasePoint,SizeDouble aMapSize,Size aBitmapSize)
        {
            Init(aBasePoint.X, aBasePoint.Y, aMapSize.Width, aMapSize.Height, aBitmapSize.Width, aBitmapSize.Height);   
        }

        public static void Zoom(int aKoef,PointDouble aBasePoint)
        {
            mPixelsPerDegree.Width+= aKoef;
            mPixelsPerDegree.Height = (int)(mPixelsPerDegree.Width * mAspectRatio);
            mBasePoint = aBasePoint;
        }

        public static void Move(PointDouble aDelta)
        {
            mBasePoint.X -= aDelta.X;
            mBasePoint.Y -= aDelta.Y;
        }

    }



    public class Geometry
    {
        
        private const double cdblRad = 6370997;
        // DistPointToPoint method
        //    calculates distance using spherical approximation
        public static double DistPointToPoint(double x1, double y1, double x2, double y2)
        {
            double tmpX1 = Math.PI * (x1 / (double)180);
            double tmpX2 = Math.PI * (x2 / (double)180);
            double tmpY1 = Math.PI * (y1 / (double)180);
            double tmpY2 = Math.PI * (y2 / (double)180);
            double newX1 = Math.Sqrt(2 * (cdblRad * Math.Cos(tmpY1)) * (cdblRad * Math.Cos(tmpY1)) * (1 - Math.Cos(tmpX1 - tmpX2)));
            double newX2 = Math.Sqrt(2 * (cdblRad * Math.Cos(tmpY2)) * (cdblRad * Math.Cos(tmpY2)) * (1 - Math.Cos(tmpX1 - tmpX2)));
            double newY = Math.Sqrt(2 * cdblRad * cdblRad * (1 - Math.Cos(tmpY1 - tmpY2)));
            double h = (newY * newY) - (((newX1 - newX2) / 2) * ((newX1 - newX2) / 2));
            double z = Math.Sqrt(h + ((newX2 - Math.Abs(newX2 - newX1) / 2) * (newX2 - Math.Abs(newX2 - newX1) / 2)));
            double c = 1 - (((z / cdblRad) * (z / cdblRad)) / 2);
            if (c == 1)
                return 0;

            c = Math.Atan((-c) / Math.Sqrt((-c) * c + 1)) + 2 * Math.Atan((double)1);
            return Math.Abs(cdblRad * c);
            
        }

        // DistPointToLine method
        //    calculates distance from a point to line using spherical approximation
        public static double DistPointToLine(double x1, double y1, double x2, double y2, double x, double y)
        {
            double ortoX, ortoY;
            double a, b, c, d;
            // calculating orto point
            if ((x1 == x2) && (y1 == y2))
            {
                ortoX = x1;
                ortoY = y1;
            }
            else
            {
                a = x2 - x1;
                b = y2 - y1;
                c = a * x + b * y;
                d = a * y1 - b * x1;
                ortoX = (a * c - b * d) / (a * a + b * b);
                ortoY = (b * c + a * d) / (a * a + b * b);
            }

            // detecting if lies on line
            double minX, maxX, minY, maxY;
            //----------------------------
            if (x1 < x2)
            {
                minX = x1;
                maxX = x2;
            }
            else
            {
                minX = x2;
                maxX = x1;
            }
            //----------------------------
            if (y1 < y2)
            {
                minY = y1;
                maxY = y2;
            }
            else
            {
                minY = y2;
                maxY = y1;
            }
            //----------------------------
            if (!((ortoX >= minX) && (ortoX <= maxX) && (ortoY >= minY) && (ortoY <= maxY)))
            {
                if (((ortoX - x1) * (ortoX - x1) + (ortoY - y1) * (ortoY - y1)) < ((ortoX - x2) * (ortoX - x2) + (ortoY - y2) * (ortoY - y2)))
                {
                    ortoX = x1;
                    ortoY = y1;
                }
                else
                {
                    ortoX = x2;
                    ortoY = y2;
                }
            }
            //----------------------------
            return DistPointToPoint(x, y, ortoX, ortoY);
        }

        public static int DistFromDegPointToDegPointInPixels(double x1, double y1, double x2, double y2,Size aPixelsPerDegree)
        {
                double delta_x = x2 - x1; delta_x = delta_x * aPixelsPerDegree.Width;
                double delta_y = y2 - y1; delta_y = delta_y * aPixelsPerDegree.Height;
                double dist_deg = Math.Sqrt(Math.Pow(delta_x, 2) + Math.Pow(delta_y, 2));
                return (int)(dist_deg);
        }

        public static int DistDegLineToDegPointInPixels(double x1, double y1, double x2, double y2, double x, double y, Size aPixelsPerDegree)
        {
            double ortoX, ortoY;
            double a, b, c, d;
            // calculating orto point
            if ((x1 == x2) && (y1 == y2))
            {
                ortoX = x1;
                ortoY = y1;
            }
            else
            {
                a = x2 - x1;
                b = y2 - y1;
                c = a * x + b * y;
                d = a * y1 - b * x1;
                ortoX = (a * c - b * d) / (a * a + b * b);
                ortoY = (b * c + a * d) / (a * a + b * b);
            }

            // detecting if lies on line
            double minX, maxX, minY, maxY;
            //----------------------------
            if (x1 < x2)
            {
                minX = x1;
                maxX = x2;
            }
            else
            {
                minX = x2;
                maxX = x1;
            }
            //----------------------------
            if (y1 < y2)
            {
                minY = y1;
                maxY = y2;
            }
            else
            {
                minY = y2;
                maxY = y1;
            }
            //----------------------------
            if (!((ortoX >= minX) && (ortoX <= maxX) && (ortoY >= minY) && (ortoY <= maxY)))
            {
                if (((ortoX - x1) * (ortoX - x1) + (ortoY - y1) * (ortoY - y1)) < ((ortoX - x2) * (ortoX - x2) + (ortoY - y2) * (ortoY - y2)))
                {
                    ortoX = x1;
                    ortoY = y1;
                }
                else
                {
                    ortoX = x2;
                    ortoY = y2;
                }
            }
            //----------------------------
            return DistFromDegPointToDegPointInPixels(x, y, ortoX, ortoY,aPixelsPerDegree);
        }

    }

    public class MyConvert
    {
        public static Color Rgb24ToColor(int aRGB)
        {
            return  
                    Color.FromArgb((int)((UInt32)aRGB | 0xFF000000));
        }
        public static int ColorToRgb24(Color aColor)
        {
            UInt32 result ;
            result = (UInt32)(aColor.R<<16);
            result = result | ((UInt32)(aColor.G << 8));
            result = result | ((UInt32)(aColor.B));
            return (int)result;   
        }


        public static bool TryParseDouble(string aStrValue, out double aValue,IFormatProvider aFormatPrivider)
        {
            
            try
            {
                aValue = double.Parse(aStrValue,aFormatPrivider);
                return true;
            }
            catch
            {
                aValue = 0;
                return false;
            }

        }
        public static bool TryParseDouble(string aStrValue, out double aValue)
        {
            return TryParseDouble(aStrValue,out aValue,Consts.nfi);
        }
        public static bool TryParseInt(string aStrValue, out int aValue, IFormatProvider aFormatPrivider)
        {

            try
            {
                aValue = int.Parse(aStrValue, aFormatPrivider);
                return true;
            }
            catch
            {
                aValue = 0;
                return false;
            }

        }
        public static bool TryParseInt(string aStrValue, out int aValue)
        {
            return TryParseInt(aStrValue, out aValue,Consts.nfi);
        }
        public static bool TryParseUInt16(string aStrValue, out UInt16 aValue, IFormatProvider aFormatPrivider)
        {

            try
            {
                aValue = UInt16.Parse(aStrValue, aFormatPrivider);
                return true;
            }
            catch
            {
                aValue = 0;
                return false;
            }

        }
        public static bool TryParseUInt16(string aStrValue, out UInt16 aValue)
        {
            return TryParseUInt16(aStrValue, out aValue,Consts.nfi);
        }
    }


    public struct Styles
    {
        public int SymbolID;
        public int SymbolSize;
        public int LineWidth;
        public int LineColor;
        public int SolidColor;
        
        public Styles(int aSymbolID, int aLineWidth, int aLineColor,int aSolidColor,int aSymbolSize)  
        {
            SymbolID        =    aSymbolID;
            SymbolSize      =    aSymbolSize;
            LineWidth       =    aLineWidth;
            LineColor       =    aLineColor;
            SolidColor      =    aSolidColor;    
        }
    }


}
