using System;
using System.IO.Ports;
using System.IO;

namespace SerialIO 
{
    public class SIO:IDisposable
    {
        public enum Source {COM,FILE}

        private string m_PortName;
        private string m_FileName;
        private Source m_Source;
        private SerialPort m_Port;
        private FileStream m_FileStream;

        public SIO(string aPortName, string aFileName, Source aSource)
        {
            m_PortName = aPortName;
            m_FileName = aFileName;
            m_Source   = aSource;
            switch (m_Source)
            {
                case Source.COM: 
                    m_Port = new SerialPort(m_PortName, 4800);
                break;
            }
        }

        private bool OpenCOM()
        {
            try 
            {
                
                m_Port.ReadTimeout = 3000;
                m_Port.Open();
                return true;
            }
            catch 
            {
                return false;
            }
        }
        private void CloseCOM()
        {
            try
            {
                m_Port.Close();
            }
            catch
            {
                return;
            }

        }
        private bool OpenFile()
        {
            try
            {
                 m_FileStream = new FileStream(m_FileName, FileMode.Open);
                return true;
            }
            catch
            {
                return false;
            }
        }
        private void CloseFile()
        {
            try
            {
                m_FileStream.Close();
                m_FileStream = null;
            }
            catch
            {
                return;
            }
        }

        public bool Open()
        {
            switch (m_Source)
            {
                case Source.COM: return OpenCOM();
                case Source.FILE: return OpenFile();
                default: return false;
            }

        }
        public int  Read(byte[] aBuf, int aOffset, int aCount)
        {
            
            
                switch (m_Source)
                {
                    case Source.COM: return m_Port.Read(aBuf, 0, aCount);
                    case Source.FILE: return m_FileStream.Read(aBuf, 0, aCount);
                    default: return 0;
                }
           
           
        }
        public void Close()
        {
            switch (m_Source)
            {
                case Source.COM: CloseCOM(); break;
                case Source.FILE: CloseFile(); break;
            }
        }

        ~SIO() 
        {
            Close();
        }
        public void Dispose()
        {
            Close();
        }
    }
       


}