using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using MIF.Primitives;

namespace MIFFile
{
    public partial class frmObjectProperties : Form
    {
        
        public frmObjectProperties()
        {
            InitializeComponent();
        }
        public frmObjectProperties(BaseObject aBaseObject)
            : this()
        {
          mObject = aBaseObject;
        }

        private BaseObject mObject;

        private void PrepareForm()
        {
            lblCaption.Text = ObjectColumn.MakeRowLabel(" | ");
            ShowData();
            InitState();
        }
        private void InitState()
        {
            pnlControl.Enabled  = true;
            lbMain.Enabled      = true;
            pnlEdits.Enabled    = false;
            if (
             (lbMain.Items.Count > 0) &&
             (lbMain.SelectedIndex<0) 
                )lbMain.SelectedIndex = 0;
             
            lbMain.Focus();
        }
        private void ActiveState()
        {
            if (lbMain.SelectedIndex < 0) return;
            pnlControl.Enabled  = false;
            lbMain.Enabled      = false;
            pnlEdits.Enabled    = true;
            tbValue.Focus();
        }
        
        private void ShowData()
        {
            try
            {
                lbMain.BeginUpdate();
                lbMain.Items.Clear();
                for (int i = 0; i < mObject.Properties.Count; i++)
                {
                    lbMain.Items.Add((mObject.Properties[i] as ObjectColumn).MakeRow(" | "));
                }
            }
            finally
            {
                lbMain.EndUpdate();
            }
        }

        private void btnEdit_Click(object sender, EventArgs e)
        {
            ActiveState();
        }

        private void btnClose_Click(object sender, EventArgs e)
        {
            Close();
        }
        private bool IsCorrectInput(ObjectColumn aOc) 
        {
            if (aOc == null) return false;
            string mif_type = "";
            int width = 0;
            int dec = 0;
            if (!aOc.GetMIFTypeAndWidth(out mif_type,out width,out dec)) return false;
            
            switch (mif_type.ToUpper())
            {
                
                case MIF.MIFTypes.Character :
                    if (!MIF.MIFTypes.IsCharacter(aOc.Value,width)) { MessageBox.Show(aOc.Type,"Error"); return false; }
                break;
                case MIF.MIFTypes.Date:
                    if (!MIF.MIFTypes.IsDate(aOc.Value)) { MessageBox.Show(aOc.Type, "Error"); return false; }
                break;
                case MIF.MIFTypes.Decimal:
                    if (!MIF.MIFTypes.IsDecimal(aOc.Value, width, dec)) { MessageBox.Show(aOc.Type, "Error"); return false; }
                break;
                case MIF.MIFTypes.Float:
                    if (!MIF.MIFTypes.IsFloat(aOc.Value)) { MessageBox.Show(aOc.Type, "Error"); return false; }
                break;
                case MIF.MIFTypes.Integer:
                    if (!MIF.MIFTypes.IsInteger(aOc.Value)) { MessageBox.Show(aOc.Type, "Error"); return false; }
                break;
                case MIF.MIFTypes.Logical:
                    if (!MIF.MIFTypes.IsLogical(aOc.Value)) { MessageBox.Show(aOc.Type, "Error"); return false; }
                break;
                case MIF.MIFTypes.SmallInteger:
                    if (!MIF.MIFTypes.IsSmallInteger(aOc.Value)) { MessageBox.Show(aOc.Type,"Error"); return false;}
                break;
                default: 
                    return false;
            }
            return true;
        }
       

        private void btnOk_Click(object sender, EventArgs e)
        {
            ObjectColumn oc = null;
            ObjectColumn tmp_oc = null;
            if (lbMain.SelectedIndex < 0) return;
            int Idx = lbMain.SelectedIndex;
            oc = (mObject.Properties[lbMain.SelectedIndex] as ObjectColumn);

            string value_text = "";

            value_text = tbValue.Text;
            
            tmp_oc = new ObjectColumn(oc.Name,value_text , oc.Type);
            if (!IsCorrectInput(tmp_oc)) { tbValue.Focus();return; }
            oc.Value=tmp_oc.Value;
            ShowData();
            lbMain.SelectedIndex = Idx;
            InitState();
        }

        private void btnCancel_Click(object sender, EventArgs e)
        {
            InitState();
        }

        private void frmObjectProperties_Shown(object sender, EventArgs e)
        {
            PrepareForm();
        }

        private void lbMain_SelectedIndexChanged(object sender, EventArgs e)
        {
            ObjectColumn oc = null;
            if (lbMain.SelectedIndex < 0) return;
            oc = (mObject.Properties[lbMain.SelectedIndex] as ObjectColumn);
            tbValue.Text = oc.Value;

        }

        public static bool Run(BaseObject aBaseObject)
        {
            using(frmObjectProperties frm = new frmObjectProperties(aBaseObject))
            {
                return frm.ShowDialog() == DialogResult.OK;
            }
        }

        private void frmObjectProperties_KeyDown(object sender, KeyEventArgs e)
        {
            switch (e.KeyCode)
            {
                case Keys.Return: if (lbMain.Focused) btnEdit_Click(btnEdit,null);
                    else
                        SelectNextControl(ActiveControl, true, true, true, true);
                break;
            case Keys.Escape: if (lbMain.Focused) btnClose_Click(btnClose,null);
                else btnCancel_Click(btnCancel,null);
                break;


                                    
            }
        }
    }
}